from __future__ import absolute_import
from .base_properties import DefaultProperty
from .utils import check_type, SimpleDict


class JsonArray(list):
    def __new__(cls, _obj=None, wrapper=None, type_config=None):
        if _obj is not None and wrapper is None and type_config is None:
            return list(_obj) if not isinstance(_obj, list) else _obj
        return super().__new__(cls, _obj, wrapper, type_config)

    def __init__(self, _obj, wrapper, type_config):
        super(JsonArray, self).__init__()
        self._obj = check_type(_obj, list,
                               'JsonArray must wrap a list or None')

        assert type_config is not None
        self._type_config = type_config
        self._wrapper = (
            wrapper or
            DefaultProperty(type_config=self._type_config)
        )
        for item in self._obj:
            super(JsonArray, self).append(self._wrapper.wrap(item))

    def __getstate__(self):
        # This breaks symmetry with JsonObject.__getstate__(), which
        # calls and returns self.to_json(). Here the JSON-y value,
        # self._obj, is not copied because it will be repopulated by
        # deepcopy/pickle list reconstruction logic, which is done
        # after obj.__setstate__(data).
        data = self.__dict__.copy()
        data["_obj"] = []
        return data

    def validate(self, required=True):
        for obj in self:
            self._wrapper.validate(obj, required=required)

    def append(self, wrapped):
        wrapped, unwrapped = self._wrapper.unwrap(wrapped)
        self._obj.append(unwrapped)
        super(JsonArray, self).append(wrapped)

    def __delitem__(self, i):
        super(JsonArray, self).__delitem__(i)
        del self._obj[i]

    def __setitem__(self, i, wrapped):
        if isinstance(i, slice):
            new_wrapped = []
            unwrapped = []
            for _wrapped in wrapped:
                _wrapped, _unwrapped = self._wrapper.unwrap(_wrapped)
                new_wrapped.append(_wrapped)
                unwrapped.append(_unwrapped)
        else:
            new_wrapped, unwrapped = self._wrapper.unwrap(wrapped)
        self._obj[i] = unwrapped
        super(JsonArray, self).__setitem__(i, new_wrapped)

    def extend(self, wrapped_list):
        if wrapped_list:
            wrapped_list, unwrapped_list = zip(
                *map(self._wrapper.unwrap, wrapped_list)
            )
        else:
            unwrapped_list = []
        self._obj.extend(unwrapped_list)
        super(JsonArray, self).extend(wrapped_list)

    def insert(self, index, wrapped):
        wrapped, unwrapped = self._wrapper.unwrap(wrapped)
        self._obj.insert(index, unwrapped)
        super(JsonArray, self).insert(index, wrapped)

    def remove(self, value):
        i = self.index(value)
        super(JsonArray, self).remove(value)
        self._obj.pop(i)

    def pop(self, index=-1):
        self._obj.pop(index)
        return super(JsonArray, self).pop(index)

    def sort(self, cmp=None, key=None, reverse=False):
        zipped = list(zip(self, self._obj))
        if key:
            new_key = lambda pair: key(pair[0])
            zipped.sort(key=new_key, reverse=reverse)
        elif cmp:
            new_cmp = lambda pair1, pair2: cmp(pair1[0], pair2[0])
            zipped.sort(cmp=new_cmp, reverse=reverse)
        else:
            zipped.sort(reverse=reverse)

        wrapped_list, unwrapped_list = list(zip(*zipped))
        while self:
            self.pop()
        super(JsonArray, self).extend(wrapped_list)
        self._obj.extend(unwrapped_list)

    def reverse(self):
        self._obj.reverse()
        super(JsonArray, self).reverse()

    def __fix_slice(self, i, j):
        length = len(self)
        if j < 0:
            j += length
        if i < 0:
            i += length
        if i > length:
            i = length
        if j > length:
            j = length
        return i, j

    def __setslice__(self, i, j, sequence):
        i, j = self.__fix_slice(i, j)
        for _ in range(j - i):
            self.pop(i)
        for k, wrapped in enumerate(sequence):
            self.insert(i + k, wrapped)

    def __delslice__(self, i, j):
        i, j = self.__fix_slice(i, j)
        for _ in range(j - i):
            self.pop(i)

    def __iadd__(self, b):
        self.extend(b)
        return self


class JsonDict(SimpleDict):

    def __init__(self, _obj=None, wrapper=None, type_config=None):
        super(JsonDict, self).__init__()
        self._obj = check_type(_obj, dict, 'JsonDict must wrap a dict or None')
        assert type_config is not None
        self._type_config = type_config
        self._wrapper = (
            wrapper or
            DefaultProperty(type_config=self._type_config)
        )
        for key, value in self._obj.items():
            self[key] = self.__wrap(key, value)

    def validate(self, required=True):
        for obj in self.values():
            self._wrapper.validate(obj, required=required)

    def __wrap(self, key, unwrapped):
        return self._wrapper.wrap(unwrapped)

    def __unwrap(self, key, wrapped):
        return self._wrapper.unwrap(wrapped)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            key = unicode(key)

        wrapped, unwrapped = self.__unwrap(key, value)
        self._obj[key] = unwrapped
        super(JsonDict, self).__setitem__(key, wrapped)

    def __delitem__(self, key):
        del self._obj[key]
        super(JsonDict, self).__delitem__(key)

    def __getitem__(self, key):
        if isinstance(key, int):
            key = unicode(key)
        return super(JsonDict, self).__getitem__(key)


class JsonSet(set):
    def __init__(self, _obj=None, wrapper=None, type_config=None):
        super(JsonSet, self).__init__()
        if isinstance(_obj, set):
            _obj = list(_obj)
        self._obj = check_type(_obj, list, 'JsonSet must wrap a list or None')
        assert type_config is not None
        self._type_config = type_config
        self._wrapper = (
            wrapper or
            DefaultProperty(type_config=self._type_config)
        )
        for item in self._obj:
            super(JsonSet, self).add(self._wrapper.wrap(item))

    def validate(self, required=True):
        for obj in self:
            self._wrapper.validate(obj, required=required)

    def add(self, wrapped):
        wrapped, unwrapped = self._wrapper.unwrap(wrapped)
        if wrapped not in self:
            self._obj.append(unwrapped)
            super(JsonSet, self).add(wrapped)

    def remove(self, wrapped):
        wrapped, unwrapped = self._wrapper.unwrap(wrapped)
        if wrapped in self:
            self._obj.remove(unwrapped)
            super(JsonSet, self).remove(wrapped)
        else:
            raise KeyError(wrapped)

    def discard(self, wrapped):
        try:
            self.remove(wrapped)
        except KeyError:
            pass

    def pop(self):
        # get first item
        for wrapped in self:
            break
        else:
            raise KeyError()
        wrapped_, unwrapped = self._wrapper.unwrap(wrapped)
        assert wrapped is wrapped_
        self.remove(unwrapped)
        return wrapped

    def clear(self):
        while self:
            self.pop()

    def __ior__(self, other):
        for wrapped in other:
            self.add(wrapped)
        return self

    def update(self, *args):
        for wrapped_list in args:
            self |= set(wrapped_list)

    union_update = update

    def __iand__(self, other):
        for wrapped in list(self):
            if wrapped not in other:
                self.remove(wrapped)
        return self

    def intersection_update(self, *args):
        for wrapped_list in args:
            self &= set(wrapped_list)

    def __isub__(self, other):
        for wrapped in list(self):
            if wrapped in other:
                self.remove(wrapped)
        return self

    def difference_update(self, *args):
        for wrapped_list in args:
            self -= set(wrapped_list)

    def __ixor__(self, other):
        removed = set()
        for wrapped in list(self):
            if wrapped in other:
                self.remove(wrapped)
                removed.add(wrapped)
        self.update(other - removed)
        return self

    def symmetric_difference_update(self, *args):
        for wrapped_list in args:
            self ^= set(wrapped_list)



# cython:profile=True
# cython:language_level=3
# -*- coding:utf-8 -*-

import logging
import random
from libc.stdlib cimport rand, RAND_MAX, srand, malloc, free
from libc.stdint cimport uintptr_t
import pandas as pd
from pandas.api.types import is_integer_dtype, is_float_dtype, is_string_dtype
import typing
from typing import TYPE_CHECKING, Type, List, Dict, Union, Set, Optional, TypeVar, Type, Generic
from .basics cimport Agent
from MelodieInfra import MelodieExceptions, MelodieException, show_prettified_warning
from collections.abc import Sequence
from cython.operator cimport dereference as deref, preincrement as inc
from cpython.ref cimport PyObject  # somewhere at the top
from cpython cimport PyObject_GetAttr, PyObject_GetAttrString, \
    PyObject_GetItem, PyList_GetItem, PyList_Size, PyObject_SetAttr, PyObject_CallFunction, PyObject_CallFunctionObjArgs
from .fastrand import sample

AgentGeneric = TypeVar('AgentGeneric')
if TYPE_CHECKING:
    from .model import Model
    from .scenario_manager import Scenario

logger = logging.getLogger(__name__)

cdef bint bs_key(Agent agent):
    return agent.id

cdef long _binary_search_agents(list lis, long num):
    cdef long left = 0
    cdef long right = len(lis) - 1
    cdef long mid = 0
    cdef long agent_id = 0
    while left <= right:
        mid = (left + right) // 2
        agent_id = bs_key(lis[mid])
        if num < agent_id:
            right = mid - 1
        elif num > agent_id:
            left = mid + 1
        else:
            return mid

    if mid == 0:
        return 0
    else:
        return -1

cdef class SeqIter:
    """
    The iterator to deal with for-loops in AgentList or other agent containers
    """

    def __init__(self, seq):
        self._seq = seq
        self._i = 0

    def __next__(self):
        if self._i >= len(self._seq):
            raise StopIteration
        next_item = self._seq[self._i]
        self._i += 1
        return next_item

cdef class DictIter:
    """
    The iterator to deal with for-loops in AgentList or other agent containers
    """

    def __init__(self, dic):
        self._dict = dic
        self._iter = iter(dic.items())

    def __next__(self):
        return next(self._iter)[1]

cdef class BaseAgentContainer():
    """
    The base class that contains agents
    """

    def __init__(self):
        self._id_offset = -1
        self.scenario: Union['Scenario', None] = None

    def new_id(self):
        """
        Create a new auto-increment ID

        :return:
        """
        self._id_offset += 1
        return self._id_offset


    def to_list(self, column_names: List[str] = None) -> List[Dict]:
        """
        Convert all agent properties to a list of dict.
        
        :param column_names:  property names
        :return: None
        """

    cpdef list init_agents(self) except *:
        """
        Initialize all agents in the container.
        
        During the initialization of agent, the ``setup()`` method of each agent will be called.
        
        :return: None
        """
        agents: List['AgentGeneric'] = [self.agent_class(self.new_id()) for i in
                                        range(self.initial_agent_num)]
        scenario = self.model.scenario
        
        for i, agent in enumerate(agents):
            agent.scenario = scenario
            agent.model = self.model
            agent.setup()
        return agents

    def _set_properties(self, props_df: pd.DataFrame):
        """
        Set parameters of all agents in current scenario from a pandas dataframe.

        :return: None
        """
        MelodieExceptions.Assertions.Type('props_df', props_df, pd.DataFrame)

        param_names = [param for param in props_df.columns if param not in
                       {'id_scenario'}]

        # props_df_cpy: Optional[pd.DataFrame] = None
        if "id_scenario" in props_df.columns:
            props_df_cpy = props_df.query(f"id_scenario == {self.scenario.id}").copy(True)
        else:
            props_df_cpy = props_df.copy()  # deep copy this dataframe.
        
        props_df_cpy.reset_index(drop=True, inplace=True)
        self.type_check(param_names, props_df_cpy)

        # Assign parameters to properties for each agent.
        for i, agent in enumerate(self):
            params = {}
            for agent_param_name in param_names:
                # .item() method was applied to convert pandas/numpy data into python-builtin types.
                item = props_df_cpy.loc[i, agent_param_name]
                if isinstance(item, str):
                    params[agent_param_name] = item
                else:
                    params[agent_param_name] = item.item()


            agent.set_params(params)

    def type_check(self, param_names: List[str], agent_params_df: pd.DataFrame):
        """
        Check if the parameters in the data frame has corresponding type with param_name 
        
        :param param_names: Parameter names to be checked
        :param agent_params_df: Dataframe for agent initial parameters.
        :return: None
        """
        dtypes = agent_params_df.dtypes
        dataframe_dtypes = {}
        for col, dtype in dtypes.items():
            if is_integer_dtype(dtype):
                dataframe_dtypes[col] = int
            elif is_float_dtype(dtype):
                dataframe_dtypes[col] = float
            elif is_string_dtype(dtype):
                dataframe_dtypes[col] = str
            else:
                show_prettified_warning(f"Cannot tell the type of column {col}.")
                dataframe_dtypes[col] = None
        for agent in self:
            for param_name in param_names:
                # param_type = type(getattr(agent, param_name))
                param_type = type(getattr(agent, param_name))
                if param_type == dataframe_dtypes[param_name] or param_type is None:
                    continue
                else:
                    raise MelodieExceptions.Data.ObjectPropertyTypeUnMatchTheDataFrameError(param_name, param_type,
                                                                                            dataframe_dtypes, agent)

cdef class AgentDict(BaseAgentContainer):
    def __init__(self, agent_class: Type[AgentGeneric], length: int, model: 'Model') -> None:
        super().__init__()
        self.scenario = model.scenario
        self.agent_class: Type[AgentGeneric] = agent_class
        self.initial_agent_num: int = length
        self.model = model
        self.agents = {agent.id: agent for agent in self.init_agents()}

    def __repr__(self):
        return f"<AgentList {self.agents}>"

    def __len__(self):
        return len(self.agents)

    def __getitem__(self, item):
        return self.agents.__getitem__(item)

    def __iter__(self):
        return DictIter(self.agents)
    
    cpdef Agent get_agent(self, long agent_id):
        return self.agents.get(agent_id)

    def add(self, agent=None, params=None):
        self._add(agent, params)

    cdef _add(self, Agent agent, dict params):
        new_id = self.new_id()
        if agent is not None:
            assert isinstance(agent, Agent)
        else:
            agent = self.agent_class(new_id)

        agent.scenario = self.model.scenario
        agent.model = self.model
        agent.setup()
        if params is not None:
            assert isinstance(params, dict)
            if params.get('id') is not None:
                show_prettified_warning(
                    f"Warning, agent 'id'  {agent.id} passed in 'params' will be **overridden** by a new id {new_id} auto generated by {self.__class__.__name__}."
                    )
            agent.set_params(params)
        agent.id = new_id        
        self.agents[agent.id] = agent
    
    cpdef remove(self, Agent agent):
        self.agents.pop(agent.id)

    def set_properties(self, props_df: pd.DataFrame):
        """
        Extract properties from a dataframe, and Each row in the dataframe represents the property of an agent.
        
        :param props_df:
        :return:
        """
        self._set_properties(props_df)

cdef class AgentList(BaseAgentContainer):
    def __init__(self, agent_class: Type[AgentGeneric], model: 'Model') -> None:
        super(AgentList, self).__init__()
        self._iter_index = 0
        self.scenario = model.scenario
        self.agent_class: Type[AgentGeneric] = agent_class
        self.initial_agent_num: int = -1
        self.model = model
        self._map = {}
        self.indices = &self._map
        self.agents: List[AgentGeneric] = []

    def setup_agents(self, agents_num: int, params_df: pd.DataFrame=None):
        """
        Setup agents with specific number, and initialize their property by a dataframe if a dataframe is passed.

        :param agents_num: A integer.
        :param params_df : A pandas dataframe whose specification is the same as the argument of ``set_properties``
        :return:
        """
        self.initial_agent_num = agents_num
        self.agents = self.init_agents()
        for i, agent in enumerate(self.agents):
            self._set_index(agent.id, i)
        if params_df is not None:
            self.set_properties(params_df)

    def _setup(self):
        self.setup()

    def setup(self):
        """
        The setup method.

        If you would like to define custom AgentList, please be sure to inherit this method.
        """
        pass

    def __repr__(self):
        return f"<AgentList {self.agents}>"

    def __len__(self):
        return len(self.agents)

    def __getitem__(self, item) -> AgentGeneric:
        return self.agents.__getitem__(item)

    def __iter__(self):
        self._iter_index = 0
        return SeqIter(self.agents)


    def random_sample(self, sample_num: int) -> List['AgentGeneric']:
        """
        Randomly sample ``sample_num`` agents from the container
        
        :param sample_num:
        :return:
        """
        return sample(self.agents, sample_num)
        

    cpdef remove(self, Agent agent) except *:
        """
        Remove an agent from the AgentList

        :param agent:
        :return:
        """
        # print(deref(self.indices))
        index = self._get_index(agent.id)
        
        self.agents.pop(index)

        for i, a in enumerate(self.agents):
            self._set_index(a.id, i)

    cdef void _set_index(self, long agent_id, long index) except *:
        # print(deref(self.indices))
        deref(self.indices)[agent_id] = index

    cdef long _get_index(self, long agent_id) except *:
        cdef cpp_map[long, long]* indices = self.indices
        # print("addr of indices:",<uintptr_t>&indices,<uintptr_t>&deref(self.indices), <uintptr_t>self.indices)
        cdef long cnt = indices.count(agent_id)
        if cnt < 1:
            raise IndexError("Index error occurred in the get_index")
        else:
            return indices.at(agent_id)
    
    cpdef list filter(self, condition) except *:
        """
        Filter agents satisfying the condition Callable[[Agent], bool]

        :return: a list of filtered agents
        """
        filtered_agents = []
        for agent in self.agents:
            if condition(agent):
                filtered_agents.append(agent)
        return filtered_agents

    def add(self, agent=None, params=None) -> None:
        """
        Add an agent onto this agent list.
        
        Notice: 
        
        1. The ``agent`` object should have the same type as this agent list's ``agent_class``.
        2. The ``id`` of new agent will be overriden by an auto-increment number.
        
        
        :param agent: If None, an agent of corresponding type will be automatically created.
        :return: None
        """
        self._add(agent, params)

    cpdef _add(self, Agent agent, dict params):
        """
        Add an agent to this AgentList

        :param agent: Optional
        :param params: Optional
        :return:
        """
        new_id = self.new_id()
        if agent is not None:
            assert isinstance(agent, Agent)
        else:
            agent = self.agent_class(new_id)

        agent.scenario = self.model.scenario
        agent.model = self.model
        agent.setup()
        if params is not None:
            assert isinstance(params, dict)
            if params.get('id') is not None:
                show_prettified_warning(
                    f"Warning, agent 'id'  {agent.id} passed in 'params' will be **overridden** by a new id {new_id} auto generated by {self.__class__.__name__}."
                    )
            agent.set_params(params)
        agent.id = new_id
        self.agents.append(agent)
        
        self._set_index(agent.id, len(self.agents) - 1)

    def to_list(self, column_names: List[str]) -> List[Dict]:
        """
        Dump all agent and their properties into a list of dict.
        
        :param column_names:  The property names to be dumped.
        :return:
        """
        
        data_list = []
        if len(self.agents) == 0:
            raise MelodieExceptions.Agents.AgentListEmpty(self)

        agent0 = self.agents[0]
        for column_name in column_names:
            if not hasattr(agent0, column_name):
                raise MelodieExceptions.Agents.AgentPropertyNameNotExist(column_name, agent0)
        for agent in self.agents:
            d = {k: getattr(agent, k) for k in column_names}
            d['id'] = agent.id
            data_list.append(d)
        return data_list

    def to_dataframe(self, column_names: List[str] = None) -> pd.DataFrame:
        """
        Store all agent values to dataframe.
        This method is always called by the data collector.

        :param column_names: property names to store
        :return:
        """
        data_list = self.to_list(column_names)
        df = pd.DataFrame(data_list)
        df['id'] = df['id'].astype(int)
        return df

    def set_properties(self, props_df: pd.DataFrame):
        """
        Extract properties from a dataframe, and Each row in the dataframe represents the property of an agent.
        
        For example, if there are 100 agents with ``id`` from 0 to 99, and each agent contains properties ``a: int`` and ``b: float``, 
        the ``props_df`` could be like this:

        
        .. code-block:: python

            import random
            import pandas as pd
            
            df = pd.DataFrame([{"id": i, "a": 2, "b": random.random()} for i in range(100)]) 
            print(df.head())
            
        The output is(df has 100 rows):

        .. code-block:: sh

                id  a         b
            0    0  2  0.207738
            1    1  2  0.236711
            2    2  2  0.869793
            3    3  2  0.797763
            4    4  2  0.900024

        For each scenario, same parameter values inside props_df will be assigned to each agents.
        As an instance, properties ``b`` value at ``agent 0`` in both ``scenario 0`` and ``1`` are both ``0.207738``.

        To set different initial parameters for different scenarios, please add a new column ``id_scenario``.
        An example of dataframe structure is shown below, and in this case, ``b`` of ``agent 0`` in ``scenario 0`` is ``0.207738``,
        while in ``scenario 1`` it was ``0.778997``

        .. code-block:: sh

                id_scenario  id  a         b
            0             0   0  2  0.207738
            1             0   1  2  0.236711
            2             0   2  2  0.869793
            ......
            100           1   0  2  0.778997
            101           1   1  2  0.450674


        :param props_df: ``pd.DataFrame`` containing agent initial properties.
        :return:
        """
        self._set_properties(props_df)
        self.agents.sort(key=lambda agent: agent.id)


    def all_agent_ids(self) -> List[int]:
        """
        Get ``id`` of all agents.

        :return: A ``list`` of ``int``.
        """
        return [agent.id for agent in self.agents]



    # cpdef get_agent(self, long agent_id):
    #     """
    #     Get an agent from the agent list

    #     :param agent_id:
    #     """
    #     index = _binary_search_agents(self.agents, agent_id)
    #     if index == -1:
    #         return None
    #     else:
    #         return self.agents[index]

    cpdef Agent get_agent(self, long agent_id):
        """
        Get an agent from the agent list. If agent unexist, return None.

        :param agent_id: The id of the agent object.
        :return: An agent object, or None.
        """
        index = self._get_index(agent_id)
        if index == -1:
            return None
        else:
            return self.agents[index]

    # @cython.nonecheck(False)
    # @cython.boundscheck(False)
    cpdef method_foreach(self, str method_name, tuple args) except *:
        """
        For each agent, execute theirs method ``method_name`` with arguments ``args``

        :param method_name: Name of method, a ``str``;
        :param args: Arguments of a method, a ``tuple``
        :return: None
        """
        method = getattr(self.agent_class, method_name)
        cdef Agent agent
        agent_num = len(self.agents) 
        cdef PyObject* ptr
        
        for i in range(agent_num):
            
            agent = <Agent>PyList_GetItem(self.agents, <Py_ssize_t>(i))
            ptr = <PyObject *>agent
            PyObject_CallFunctionObjArgs(method, ptr, NULL)
            # method(agent, *args)

    cpdef vectorize(self, str prop_name) except *:
        """
        NotImplemented yet.

        Generate an numpy array from this list, where the values come from the property defined by ``prop_name`` on each agent.

        :param prop_name: Property name
        :return: An 1-D Numpy array
        """
        if len(self.agents)==0:
            return
        raise NotImplementedError

cdef class Container:
    cdef PyObject ** ptr
    cdef Agent agent
    cdef long size
    cdef list mgr_list
    def __init__(self, size):
        self.ptr = <PyObject **> malloc(size * sizeof(PyObject*))
        self.mgr_list = []
        self.size = size
        for i in range(size):
            agent = Agent(i)
            self.mgr_list.append(agent)
            self.ptr[i] = <PyObject *>agent

        # print(sizeof(Agent), sizeof(PyObject))
        # for i in range(size):
    cdef walk1(self) except *:
        cdef long s = 0
        for i in range(self.size):
            self.agent = <Agent>self.ptr[i]
            s+=self.agent.id
            # print("agent_id", agent.id)

    cdef walk2(self) except *:
        cdef long s = 0
        cdef long a = 0
        for i in range(self.size):
            a+=5*self.size
            self.agent = <Agent>self.mgr_list[i]
            s+=self.agent.id

    def __del__(self):
        free(self.ptr)

cpdef test_container() except *:
    import time
    c = Container(10000)
    M = 10000
    t1 = time.time()
    for i in range(M):
        c.walk1()
    t2 = time.time()
    for i in range(M):
        c.walk2()    
    t3 = time.time()
    print("walk1", t2-t1, 'walk2', t3-t2)
    


from typing import Any, Optional, TYPE_CHECKING, Dict, List

import pandas as pd

if TYPE_CHECKING:
    from Melodie.model import Model
    from Melodie.scenario_manager import Scenario

class Element:
    def set_params(self, params: Dict[str, Any]): ...

class Agent(Element):
    def __init__(self, agent_id: int):
        self.id = agent_id
        self.scenario: Optional["Scenario"] = None
        self.model: Optional["Model"] = None
    def setup(self):
        pass
    def __repr__(self) -> str: ...
    def set_params(self, params: Dict[str, Any]): ...
    def get_style(self): ...
    def to_dict(self, attributes: List[str]): ...

class Environment(Element):
    def __init__(self):
        self.model: Optional["Model"] = None
        self.scenario: Optional["Scenario"] = None
    def setup(self) -> None: ...
    def to_dict(self, properties: List[str]) -> Dict: ...
    def to_dataframe(self, properties: List[str]) -> pd.DataFrame: ...
    def _setup(self) -> None: ...



# cython:language_level=3
# -*- coding:utf-8 -*-

import functools
from typing import Type, Set, Dict, List, Tuple
from cpython.ref cimport PyObject  # somewhere at the top
from cpython cimport PyObject_GetAttr, PyObject_GetAttrString, \
    PyObject_GetItem, PyList_GetItem, PyList_Size, PyObject_SetAttr
cimport cython
from libc.stdlib cimport rand, RAND_MAX, srand, malloc, free
cimport numpy as np
import numpy as np

from libc.math cimport sin, cos, sqrt, log, ceil
import random

from math import log as _log, exp as _exp, pi as _pi, e as _e, ceil as _ceil
from math import sqrt as _sqrt, acos as _acos, cos as _cos, sin as _sin
from os import urandom as _urandom

import random
import time
cimport numpy as np
import numpy as np
from cpython cimport array
import array

ctypedef np.int64_t DTYPE_t

srand(time.time())

cdef double uniform():
    return (rand()/(RAND_MAX*1.0))

@cython.cdivision(True)
cdef long randint(long _min, long _max):
    return  rand() % (_max-_min+1) + _min

cdef class ReservoirSample(object):
    cdef int _size
    cdef int _counter
    cdef long *_sample

    cdef void _initialize(self, int size):
        self._size = size
        self._counter = 0
        self._sample = <long *> malloc(size * sizeof(long))


    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void feed(self, long item):
        cdef long rand_int
        self._counter += 1
        # The i.th element(i <= k) directly put into the pool
        if self._counter-1 < self._size:
            self._sample[self._counter-1] = item
            return
        # The i.th element (i > k) put into the pool with the possibility of k/i
        rand_int = randint(1, self._counter)
        if rand_int <= self._size:
            self._sample[rand_int - 1] = item

    cdef long* _get_samples(self):
        return self._sample


@cython.nonecheck(False)
cdef long* _reservoir_sample(int size, int num):
    """

    Return the pointer of reservoir array
    """
    cdef list sample
    cdef ReservoirSample rs
    cdef int item
    rs = ReservoirSample()
    rs._initialize(num)
    for item in range(size):
        rs.feed(item)
    return rs._get_samples()


cpdef list reservoir_sample(list population, int num):
    cdef int size = len(population)

    cdef long sample_index
    cdef long* samples = _reservoir_sample(size, num)
    cdef list result = [None] * num
    for i in range(num):
        sample_index = samples[i]
        result[i] = population[sample_index]
    free(samples)
        
    return result 

@cython.cdivision(True)
cdef double log_base(double x, double base):
    return log(x)/log(base)

cpdef object sample(list population, int sample_num):
    """
    Randomly sample `sample_num` agents from the container
    This code is the copied from standard library `random` and did some performance optimizations. 

    :param sample_num:
    :return:
    """
    
    cdef int k = sample_num
    cdef int n = len(population)
    cdef int setsize = 21        # size of a small set minus size of an empty list
    cdef list result
    cdef list pool  
    cdef int i, j

    # randbelow = random._inst._randbelow

    if not 0 <= k <= n:
        raise ValueError("Sample larger than population or is negative")
    result = [None] * k
    
    if k > 5:
        # setsize += 4 ** _ceil(_log(k * 3, 4)) # table size for big sets
        setsize += 4 ** <int>ceil(log_base(k * 3, 4)) # table size for big sets
    # print(n<=setsize)
    if n <= setsize:
        # # An n-length list is smaller than a k-length set
        # pool = list(population)
        # for i in range(k):         # invariant:  non-selected at [0,n-i)
        #     j = randbelow(n-i)
        #     result[i] = pool[j]
        #     pool[j] = pool[n-i-1]   # move non-selected item into vacancy
        return reservoir_sample(population, sample_num)
    else:
        selected = set()
        selected_add = selected.add
        for i in range(k):
            # j = randbelow(n)
            j = randint(0, n-1)
            while j in selected:
                # j = randbelow(n)
                j = randint(0, n-1)
            selected_add(j)
            # print(j)
            result[i] = population[j]
    return result


from .actions import Action
from .params import FloatParam, ArrayParam, IntParam, Param, BoolParam, StringParam
from .vis_charts import JSONBase
from .vis_agent_series import AgentSeriesManager, AgentSeries
from .visualizer import BaseVisualizer, NetworkVisualizer, Visualizer, MelodieModelReset
from .ws_protocol import create_file_response, create_failed_response



from __future__ import absolute_import
import six
import inspect
from .exceptions import BadValueError

function_name = None
if six.PY3:
    def function_name(f):
        return f.__name__
else:
    def function_name(f):
        return f.func_name


class JsonProperty(object):

    default = None
    type_config = None

    def __init__(self, default=Ellipsis, name=None, choices=None,
                 required=False, exclude_if_none=False, validators=None,
                 verbose_name=None, type_config=None):
        validators = validators or ()
        self.name = name
        if default is Ellipsis:
            default = self.default
        if callable(default):
            self.default = default
        else:
            self.default = lambda: default
        self.choices = choices
        self.choice_keys = []
        if choices:
            for choice in choices:
                if isinstance(choice, tuple):
                    choice, _ = choice
                self.choice_keys.append(choice)
        self.required = required
        self.exclude_if_none = exclude_if_none
        self._validators = validators
        self.verbose_name = verbose_name
        if type_config:
            self.type_config = type_config

    def init_property(self, default_name, type_config):
        self.name = self.name or default_name
        self.type_config = self.type_config or type_config

    def wrap(self, obj):
        raise NotImplementedError()

    def unwrap(self, obj):
        """
        must return tuple of (wrapped, unwrapped)

        If obj is already a fully wrapped object,
        it must be returned as the first element.

        For an example where the first element is relevant see ListProperty

        """
        raise NotImplementedError()

    def to_json(self, value):
        _, unwrapped = self.unwrap(value)
        return unwrapped

    def to_python(self, value):
        return self.wrap(value)

    def __get__(self, instance, owner):
        if instance:
            assert self.name in instance
            return instance[self.name]
        else:
            return self

    def __set__(self, instance, value):
        instance[self.name] = value

    def __call__(self, method):
        """
        use a property as a decorator to set its default value

        class Document(JsonObject):
            @StringProperty()
            def doc_type(self):
                return self.__class__.__name__
        """
        assert self.default() is None
        self.default = method
        self.name = self.name or function_name(method)
        return self

    def exclude(self, value):
        return self.exclude_if_none and not value

    def empty(self, value):
        return value is None

    def validate(self, value, required=True, recursive=True):
        if (self.choice_keys and value not in self.choice_keys
                and value is not None):
            raise BadValueError(
                '{0!r} not in choices: {1!r}'.format(value, self.choice_keys)
            )

        if not self.empty(value):
            self._custom_validate(value)
        elif required and self.required:
            raise BadValueError(
                'Property {0} is required.'.format(self.name)
            )
        if recursive and hasattr(value, 'validate'):
            value.validate(required=required)

    def _custom_validate(self, value):
        if self._validators:
            if hasattr(self._validators, '__iter__'):
                for validator in self._validators:
                    validator(value)
            else:
                self._validators(value)


class JsonContainerProperty(JsonProperty):

    _type = default = None
    container_class = None

    def __init__(self, item_type=None, **kwargs):
        self._item_type_deferred = item_type
        super(JsonContainerProperty, self).__init__(**kwargs)

    def init_property(self, **kwargs):
        super(JsonContainerProperty, self).init_property(**kwargs)
        if not inspect.isfunction(self._item_type_deferred):
            # trigger validation
            self.item_wrapper

    def to_item_wrapper(self, item_type):
        from .base import JsonObjectMeta
        from .properties import ObjectProperty
        if item_type is None:
            return None
        if isinstance(item_type, JsonObjectMeta):
            return ObjectProperty(item_type, type_config=self.type_config)
        elif isinstance(item_type, JsonProperty):
            item_wrapper = item_type
            if item_wrapper.type_config is None:
                item_wrapper.type_config = self.type_config
            return item_wrapper
        elif issubclass(item_type, JsonProperty):
            return item_type(type_config=self.type_config, required=True)
        elif item_type in self.type_config.properties:
            return self.type_config.properties[item_type](type_config=self.type_config, required=True)
        else:
            for general_type, property_cls in self.type_config.properties.items():
                if issubclass(item_type, general_type):
                    return property_cls(type_config=self.type_config, required=True)
            raise ValueError("item_type {0!r} not in {1!r}".format(
                item_type,
                self.type_config.properties,
            ))

    @property
    def item_wrapper(self):
        if hasattr(self, '_item_type_deferred'):
            if inspect.isfunction(self._item_type_deferred):
                self._item_wrapper = self.to_item_wrapper(self._item_type_deferred())
            else:
                self._item_wrapper = self.to_item_wrapper(self._item_type_deferred)
            del self._item_type_deferred
        return self._item_wrapper

    def empty(self, value):
        return not value

    def wrap(self, obj):
        return self.container_class(obj, wrapper=self.item_wrapper,
                                    type_config=self.type_config)

    def unwrap(self, obj):
        if not isinstance(obj, self._type):
            raise BadValueError(
                '{0!r} is not an instance of {1!r}'.format(
                    obj, self._type.__name__)
            )
        if isinstance(obj, self.container_class):
            return obj, obj._obj
        else:
            wrapped = self.wrap(self._type())
            self._update(wrapped, obj)
            return self.unwrap(wrapped)

    def _update(self, container, extension):
        raise NotImplementedError()


class DefaultProperty(JsonProperty):

    def wrap(self, obj):
        assert self.type_config.string_conversions is not None
        value = self.value_to_python(obj)
        property_ = self.value_to_property(value)

        if property_:
            return property_.wrap(obj)

    def unwrap(self, obj):
        property_ = self.value_to_property(obj)
        if property_:
            return property_.unwrap(obj)
        else:
            return obj, None

    def value_to_property(self, value):
        map_types_properties = self.type_config.properties
        if value is None:
            return None
        elif type(value) in map_types_properties:
            return map_types_properties[type(value)](
                type_config=self.type_config)
        else:
            for value_type, prop_class in map_types_properties.items():
                if isinstance(value, value_type):
                    return prop_class(type_config=self.type_config)
            else:
                raise BadValueError(
                    'value {0!r} not in allowed types: {1!r}'.format(
                        value, map_types_properties.keys())
                )

    def value_to_python(self, value):
        """
        convert encoded string values to the proper python type

        ex:
        >>> DefaultProperty().value_to_python('2013-10-09T10:05:51Z')
        datetime.datetime(2013, 10, 9, 10, 5, 51)

        other values will be passed through unmodified
        Note: containers' items are NOT recursively converted

        """
        if isinstance(value, six.string_types):
            convert = None
            for pattern, _convert in self.type_config.string_conversions:
                if pattern.match(value):
                    convert = _convert
                    break

            if convert is not None:
                try:
                    #sometimes regex fail so return value
                    value = convert(value)
                except Exception:
                    pass
        return value


class AssertTypeProperty(JsonProperty):
    _type = None

    def assert_type(self, obj):
        if obj is None:
            return
        elif not isinstance(obj, self._type):
            raise BadValueError(
                '{0!r} not of type {1!r}'.format(obj, self._type)
            )

    def selective_coerce(self, obj):
        return obj

    def wrap(self, obj):
        obj = self.selective_coerce(obj)
        self.assert_type(obj)
        return obj

    def unwrap(self, obj):
        obj = self.selective_coerce(obj)
        self.assert_type(obj)
        return obj, obj


class AbstractDateProperty(JsonProperty):

    _type = None

    def __init__(self, exact=False, *args, **kwargs):
        super(AbstractDateProperty, self).__init__(*args, **kwargs)
        self.exact = exact

    def wrap(self, obj):
        try:
            if not isinstance(obj, six.string_types):
                raise ValueError()
            return self._wrap(obj)
        except ValueError:
            raise BadValueError('{0!r} is not a {1}-formatted string'.format(
                obj,
                self._type.__name__,
            ))

    def unwrap(self, obj):
        if not isinstance(obj, self._type):
            raise BadValueError('{0!r} is not a {1} object'.format(
                obj,
                self._type.__name__,
            ))
        return self._unwrap(obj)

    def _wrap(self, obj):
        raise NotImplementedError()

    def _unwrap(self, obj):
        raise NotImplementedError()



import logging
from typing import Callable

from MelodieInfra.exceptions.exceptions import MelodieExceptions

logger = logging.getLogger(__name__)


def args_check(func: Callable, expected_arg_num: int):
    if func.__code__.co_argcount != expected_arg_num:
        raise MelodieExceptions.Program.Function.FunctionArgsNumError(
            func, expected_arg_num, func.__code__.co_argcount
        )



class MelodieGlobalConfig:
    class Logger:
        TIME_ROUND_DIGITS = 3

        @staticmethod
        def round_elapsed_time(t: float):
            return round(t, MelodieGlobalConfig.Logger.TIME_ROUND_DIGITS)



import base64
import importlib
import json
import multiprocessing
import queue
import sys
import time
from typing import Dict, Tuple, Any

import cloudpickle

from Melodie.global_configs import MelodieGlobalConfig
from Melodie.utils.system_info import is_windows

params_queue = multiprocessing.Queue() if not is_windows() else queue.Queue()
result_queue = multiprocessing.Queue() if not is_windows() else queue.Queue()



import os
from typing import Optional, Dict, List, Union, Callable, Type

import numpy as np
import pandas as pd
import sqlalchemy

from MelodieInfra import DBConn, create_db_conn, MelodieExceptions, Config

from .scenario_manager import Scenario
from .table_generator import DataFrameGenerator

class DataFrameInfo:
    df_name: str
    columns: Dict[str, "sqlalchemy.types"]
    file_name: Optional[str] = None

    def __init__(
            self,
            df_name: str,
            columns: Dict[str, "sqlalchemy.types"],
            file_name: Optional[str] = None,
    ):
        self.df_name: str = df_name
        self.columns: Dict[str, "sqlalchemy.types"] = columns
        self.file_name: Optional[str] = file_name

    def check_column_names(self, columns: List[str]):
        if set(columns) != set(self.columns.keys()):
            missing = set(self.columns.keys()).difference(set(columns))
            undefined = set(set(columns)).difference(self.columns.keys())
            raise MelodieExceptions.Data.ColumnNameConsistencyError(
                self.df_name, missing, undefined
            )


class MatrixInfo:
    def __init__(
            self,
            mat_name: str,
            data_type: sqlalchemy.types,
            file_name: Optional[str] = None,
    ):
        self.mat_name: str = mat_name
        self.data_type: sqlalchemy.types = data_type
        self.file_name: Optional[str] = file_name

    @property
    def dtype(self):
        py_type = self.data_type.python_type
        if issubclass(py_type, int):
            return np.int
        elif issubclass(py_type, float):
            return np.float
        else:
            raise NotImplementedError(
                f"Cannot convert this type {self.data_type} to numpy data type!"
            )


class DataLoader:
    """
    DataLoader loads dataframes or matrices.
    
    ``Simulator``/``Trainer``/``Calibrator`` will have reference to DataLoader to avoid defining tables multiple times.
    """

    def __init__(
            self,
            manager,
            config: Config,
            scenario_cls: Type[Scenario],
            as_sub_worker=False,
    ):
        MelodieExceptions.Assertions.NotNone(
            scenario_cls, "Scenario class should not be None!"
        )
        self.as_sub_worker = as_sub_worker  # If loader is loaded from sub worker.
        self.config: Config = config
        self.scenario_cls = scenario_cls
        self.registered_dataframes: Optional[Dict[str, pd.DataFrame]] = {}
        self.registered_matrices: Optional[Dict[str, np.ndarray]] = {}
        self.manager = manager
        self.manager.data_loader = self
        self.setup()

    def setup(self):
        pass

    def register_dataframe(
            self, table_name: str, data_frame: pd.DataFrame, data_types: dict = None
    ) -> None:
        """
        Register a pandas dataframe.

        :param table_name: Name of dataframe
        :param data_frame: A pandas dataframe
        :param data_types: A dictionary describing data types.
        :return: None
        """
        if data_types is None:
            data_types = {}
        if not self.as_sub_worker:
            DBConn.register_dtypes(table_name, data_types)
            create_db_conn(self.config).write_dataframe(
                table_name, data_frame, data_types=data_types, if_exists="replace"
            )
        self.registered_dataframes[table_name] = create_db_conn(
            self.config
        ).read_dataframe(table_name)

    def load_dataframe(self, df_info: "DataFrameInfo") -> None:
        """
        Register static table. The static table will be copied into database.

        The scenarios/agents parameter tables can also be registered by this method.

        :return: None
        """
        table_name = df_info.df_name
        file_name = df_info.file_name
        data_types = df_info.columns
        _, ext = os.path.splitext(file_name)
        table: Optional[pd.DataFrame]

        MelodieExceptions.Data.TableNameInvalid(table_name)
        if ext in {".xls", ".xlsx"}:
            file_path_abs = os.path.join(self.config.input_folder, file_name)
            table = pd.read_excel(file_path_abs)
            df_info.check_column_names(list(table.columns))
        else:
            raise NotImplemented(file_name)
        if not self.as_sub_worker:
            DBConn.register_dtypes(table_name, data_types)
            create_db_conn(self.config).write_dataframe(
                table_name,
                table,
                data_types=data_types,
                if_exists="replace",
            )

        self.registered_dataframes[table_name] = create_db_conn(
            self.config
        ).read_dataframe(table_name)

    def load_matrix(self, matrix_info: "MatrixInfo") -> None:
        """
        Register static matrix.

        :return: None
        """
        _, ext = os.path.splitext(matrix_info.file_name)
        if ext in {".xls", ".xlsx"}:
            file_path_abs = os.path.join(
                self.config.input_folder, matrix_info.file_name
            )
            table: pd.DataFrame = pd.read_excel(file_path_abs, header=None)
            array = table.to_numpy(matrix_info.dtype, copy=True)
        else:
            raise NotImplementedError(f"Cannot load file to matrix {ext}")
        self.registered_matrices[matrix_info.mat_name] = array

    def dataframe_generator(
            self,
            df_info: DataFrameInfo,
            rows_in_scenario: Union[int, Callable[[Scenario], int]],
    ) -> DataFrameGenerator:
        """
        Create a new generator for dataframes.

        :param df_info: Dataframe info indicating the dataframe to be generated.
        :param rows_in_scenario: How many rows will be generated for a specific scenario. \
            This argument should be an integer as number of rows for each scenario, or a function with a parameter \
            with type `Scenario` and return an integer for how many rows to generate for this scenario .
        :return: A dataframe generator object
        """
        return DataFrameGenerator(self, df_info, rows_in_scenario)

    def generate_scenarios_from_dataframe(self, df_name: str) -> List["Scenario"]:
        """
        Generate scenario objects by the parameter from static table named ``df_name``.

        :param df_name: Name of static table.
        :return: A list of scenario object.
        """
        scenarios_dataframe = self.registered_dataframes.get(df_name)
        if scenarios_dataframe is None:
            MelodieExceptions.Data.TableNotFound(df_name, self.registered_dataframes)

        cols = [col for col in scenarios_dataframe.columns]
        scenarios: List[Scenario] = []
        for i in range(scenarios_dataframe.shape[0]):
            scenario = self.scenario_cls()
            scenario.manager = self.manager
            scenario.setup()

            for col_name in cols:
                value = scenarios_dataframe.loc[i, col_name]
                if isinstance(value, str):
                    scenario.__dict__[col_name] = value
                else:
                    scenario.__dict__[col_name] = value.item()
            scenarios.append(scenario)
        if len(scenarios) == 0:
            raise MelodieExceptions.Scenario.NoValidScenarioGenerated(scenarios)
        return scenarios

    def generate_scenarios(self, manager_type: str) -> List["Scenario"]:
        """
        Generate scenario objects by the parameter from static tables or scenarios_dataframe.

        :param manager_type: The type of scenario manager, a ``str`` in "simulator", "trainer" or "calibrator".
        :return: A list of scenarios.
        """
        if manager_type not in {"simulator", "trainer", "calibrator"}:
            MelodieExceptions.Program.Variable.VariableNotInSet(
                "manager_type", manager_type, {"simulator", "trainer", "calibrator"}
            )
        return self.generate_scenarios_from_dataframe(f"{manager_type}_scenarios")



from typing import Dict, List, Callable, Any

from .vis_charts import JSONBase
from .. import Agent


class AgentSeries(JSONBase):
    def __init__(
        self,
        name: str,
        type: str,
        role_getter,
        roles_repr,
        color: str = "",
        symbol="rect",
    ):
        self.name: str = name
        self.type: str = type
        self.data: List = []
        self._role_getter: Callable[["Agent"], int] = role_getter
        self._roles_repr: Dict[int, Dict[str, Any]] = roles_repr
        self.itemStyle: Dict = {
            "color": color,
        }
        self.symbol: str = symbol


class AgentSeriesManager(JSONBase):
    def __init__(self):
        self.agent_series: Dict[str, AgentSeries] = {}

    def add_series(
        self,
        series_name,
        series_type: str,
        role_getter: Callable[["Agent"], int],
        roles_repr,
        color: str = "",
        symbol: str = "",
    ):
        self.agent_series[series_name] = AgentSeries(
            series_name, series_type, role_getter, roles_repr
        )

    def set_series_data(self, series_name: str, data):
        self.agent_series[series_name].data = data

    def to_dict(self):
        return self.to_json()["agent_series"]



from __future__ import absolute_import

import datetime
import decimal
import re

import six


from .base import JsonObjectBase, _LimitedDictInterfaceMixin
from . import properties
from .containers import JsonArray, JsonDict, JsonSet


re_date = re.compile(r'^(\d{4})\D?(0[1-9]|1[0-2])\D?([12]\d|0[1-9]|3[01])$')
re_time = re.compile(
    r'^([01]\d|2[0-3])\D?([0-5]\d)\D?([0-5]\d)?\D?(\d{3,6})?$')
re_datetime = re.compile(
    r'^(\d{4})\D?(0[1-9]|1[0-2])\D?([12]\d|0[1-9]|3[01])'
    r'(\D?([01]\d|2[0-3])\D?([0-5]\d)\D?([0-5]\d)?\D?(\d{3,6})?'
    r'([zZ]|([\+-])([01]\d|2[0-3])\D?([0-5]\d)?)?)?$'
)
re_decimal = re.compile(r'^(\d+)\.(\d+)$')

if six.PY3:
    unicode = str
    long = int


class JsonObject(JsonObjectBase, _LimitedDictInterfaceMixin):
    def __getstate__(self):
        return self.to_json()

    def __setstate__(self, dct):
        self.__init__(dct)

    class Meta(object):
        properties = {
            decimal.Decimal: properties.DecimalProperty,
            datetime.datetime: properties.DateTimeProperty,
            datetime.date: properties.DateProperty,
            datetime.time: properties.TimeProperty,
            str: properties.StringProperty,
            unicode: properties.StringProperty,
            bool: properties.BooleanProperty,
            int: properties.IntegerProperty,
            long: properties.IntegerProperty,
            float: properties.FloatProperty,
            list: properties.ListProperty,
            dict: properties.DictProperty,
            set: properties.SetProperty,
            JsonArray: properties.ListProperty,
            JsonDict: properties.DictProperty,
            JsonSet: properties.SetProperty,
        }
        string_conversions = (
            (re_date, datetime.date),
            (re_time, datetime.time),
            (re_datetime, datetime.datetime),
            (re_decimal, decimal.Decimal),
        )



from Melodie import Model
from .data_collector import _ALIAS_DataCollector
from .environment import _ALIAS_Environment


class _ALIAS_Model(Model):
    def setup(self):
        with self.define_basic_components():
            self.data_collector = _ALIAS_DataCollector()
            self.environment = _ALIAS_Environment()

    def run(self):
        for t in range(0, self.scenario.period_num):
            self.data_collector.collect(t)
        self.data_collector.save()



"""
Write prettified warnings to terminal for better focus attraction.

# https://pypi.org/project/termcolor/
"""

from typing import Dict, Set

try:
    from termcolor import cprint
except:

    def cprint(s, **kwargs):
        print(s)


class FSM:
    def __init__(self, initial_state: int, state_transition_table: Dict[int, Set[int]]):
        self.current_state: int = initial_state
        self.stt = state_transition_table

    def state_change(self, new_state: int):
        assert new_state in self.stt.keys()
        assert new_state in self.stt[self.current_state], (
            new_state,
            self.current_state,
            self.stt[self.current_state],
        )
        self.current_state = new_state


class ColorParseFSM(FSM):
    NORMAL = 0
    BOLD = 1
    UNDERLINE = 2
    REVERSE = 3
    SINGLE_STAR_FROM_NORMAL = 4
    SINGLE_STAR_FROM_BOLD = 5

    transitions = {
        NORMAL: {SINGLE_STAR_FROM_NORMAL, NORMAL, UNDERLINE, REVERSE},
        BOLD: {BOLD, NORMAL, SINGLE_STAR_FROM_BOLD},
        SINGLE_STAR_FROM_NORMAL: {BOLD, UNDERLINE},
        SINGLE_STAR_FROM_BOLD: {NORMAL},
        UNDERLINE: {UNDERLINE, NORMAL},
        REVERSE: {NORMAL, REVERSE},
    }

    def __init__(self):
        super().__init__(self.NORMAL, self.transitions)

    def on_star(self):
        d = {
            self.NORMAL: self.SINGLE_STAR_FROM_NORMAL,
            self.SINGLE_STAR_FROM_NORMAL: self.BOLD,
            self.BOLD: self.SINGLE_STAR_FROM_BOLD,
            self.SINGLE_STAR_FROM_BOLD: self.NORMAL,
            self.UNDERLINE: self.NORMAL,
        }
        self.state_change(d[self.current_state])

    def on_back_quote(self):
        d = {self.NORMAL: self.REVERSE, self.REVERSE: self.NORMAL}
        self.state_change(d[self.current_state])

    def on_char(self):
        d = {
            self.SINGLE_STAR_FROM_NORMAL: self.UNDERLINE,
            self.NORMAL: self.NORMAL,
            self.BOLD: self.BOLD,
            self.UNDERLINE: self.UNDERLINE,
            self.REVERSE: self.REVERSE,
        }
        self.state_change(d[self.current_state])

    def parse(self, content: str):
        font_color = "magenta"
        on_color = None
        for i, char in enumerate(content):
            if char == "*":
                self.on_star()
                continue
            elif char == "`":
                self.on_back_quote()
                continue
            else:
                self.on_char()
            if self.current_state == self.UNDERLINE:
                # cprint(char, on_color='on_yellow', attrs=['underline'], end="")
                cprint(
                    char,
                    color=font_color,
                    on_color=on_color,
                    attrs=["underline"],
                    end="",
                )
            elif self.current_state == self.BOLD:
                cprint(
                    char, color=font_color, on_color=on_color, attrs=["bold"], end=""
                )
            elif self.current_state == self.REVERSE:
                cprint(
                    char, color=font_color, on_color=on_color, attrs=["reverse"], end=""
                )
            else:
                cprint(char, color=font_color, on_color=on_color, end="")
        print("\n", end="")


import inspect


def show_link(stackdepth=1):
    """
    Print a link in PyCharm to a line in file.

    Thanks to: https://stackoverflow.com/questions/26300594/print-code-link-into-pycharms-console
    """

    file = inspect.stack()[stackdepth].filename
    line = inspect.stack()[stackdepth].lineno

    string = f'File "{file}", line {max(line, 1)}'.replace("\\", "/")
    return string


def show_prettified_warning(warning: str):
    ColorParseFSM().parse("WARNING: " + warning)



# -*- coding:utf-8 -*-
# @Time: 2022/11/23 21:56
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: excel.py
import json
import os
import tempfile
from typing import Any, List, cast

import openpyxl
import pandas as pd

from MelodieInfra.models import ExcelWriteRequest, DataServiceStatus, ExcelReadSheetRequest, ExcelReadSheetResponse, \
    DataServiceState


def df_to_json(df: pd.DataFrame):
    with tempfile.NamedTemporaryFile(delete=False) as tf:
        df.to_json(tf.name, orient="table", indent=4, index=False)
        data = tf.read()
        tf.close()
    data = json.loads(data)
    for item in data["schema"]["fields"]:
        if item["type"] == "integer":
            item["type"] = "number"
    return data


class ExcelManipulator:
    def __init__(self, filename: str) -> None:
        self.filename = filename

    def get_sheet_names(self):
        return pd.ExcelFile(self.filename).sheet_names

    def read_sheet(self, sheet_name: str = None, **kwargs):
        return pd.read_excel(self.filename, sheet_name, **kwargs)

    def write_to_sheet(self, df: pd.DataFrame, sheet_name: str, **kwargs):
        book = None
        sheet_exist = False
        if os.path.exists(self.filename):
            book = openpyxl.load_workbook(self.filename)
            sheet_exist = sheet_name in self.get_sheet_names()
        if sheet_exist:
            with pd.ExcelWriter(
                    self.filename, engine="openpyxl", mode="a", if_sheet_exists="replace"
            ) as writer:
                df.to_excel(writer, sheet_name=sheet_name, index=False)
        else:
            writer = pd.ExcelWriter(self.filename, engine="openpyxl")
            print(sheet_exist)
            if book is not None:
                writer.book = book
            df.to_excel(writer, sheet_name, index=False)
            writer.save()


class ExcelDataService:
    @staticmethod
    def write_excel(req: ExcelWriteRequest) -> DataServiceStatus:
        """
        Write data to excel file.

        :param req:
        :return: error message, None means success.
        """
        _, ext = os.path.splitext(req.path)
        ext = ext[1:]
        if ext in {"xls", "xlsx"}:
            df = pd.DataFrame(req.data)
            try:
                if req.sheet is not None:
                    em = ExcelManipulator(req.path)
                    em.write_to_sheet(df, req.sheet)
                else:
                    df.to_excel(req.path, index=False)
                return DataServiceStatus(status=DataServiceState.SUCCESS, msg='Succeeded saved table file!')
            except BaseException as e:
                return DataServiceStatus(status=DataServiceState.ERROR, msg=f"{e}")
        else:
            return DataServiceStatus(status=DataServiceState.ERROR,
                                     msg=f"Extension name {ext} unsupported for excel files!")

    @staticmethod
    def read_excel(req: ExcelReadSheetRequest) -> DataServiceStatus:
        """
        Read one sheet from an excel file.

        :param req:
        :return:
        """
        _, ext = os.path.splitext(req.path)
        ext = ext[1:]
        if ext in {"xls", "xlsx"}:
            em = ExcelManipulator(req.path)
            sheets: List[str] = cast(Any, em.get_sheet_names())
            currentSheet: str = sheets[0] if req.sheet is None else req.sheet
            res = em.read_sheet(currentSheet)
            resp = ExcelReadSheetResponse.create(df_to_json(res), currentSheet, sheets)
            return DataServiceStatus(status=DataServiceState.SUCCESS, msg="Read excel file succeeded!", data=resp)
        else:
            return DataServiceStatus(status=DataServiceState.ERROR,
                                     msg=f"File extension {ext} unsupported for excel files!")



# -*- coding:utf-8 -*-
# @Time: 2022/11/24 17:59
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: base.py.py
from ...jsonobject import JsonObject, StringProperty, ListProperty


def assert_is_list(obj):
    assert isinstance(obj, list)


class ExcelMeta(JsonObject):
    widget = StringProperty(name="widget")
    type = StringProperty(name="type")
    sheet_names = ListProperty(str, name="sheetNames", validators=assert_is_list)
    current_sheet = StringProperty(name='currentSheet')



import time
from .intmap cimport mld_int_map_create, mld_int_map_free, IntMap, mld_int_map_get, mld_key_in_int_map


cpdef main():
    cdef IntMap* map
    cdef long i = 123
    map = mld_int_map_create(100)
    map.ptr[0] = 123
    t0 = time.time()
    mld_int_map_set(map, 0, i)
    mld_int_map_get(map, 0)
    t1 = time.time()
    print(t1-t0, mld_int_map_get(map, 0))


from __future__ import absolute_import
from __future__ import unicode_literals
from .base import JsonObjectMeta
from .containers import JsonArray
from .properties import *
from .api import JsonObject
import six

if six.PY3:
    __all__ = [
        'IntegerProperty', 'FloatProperty', 'DecimalProperty',
        'StringProperty', 'BooleanProperty',
        'DateProperty', 'DateTimeProperty', 'TimeProperty',
        'ObjectProperty', 'ListProperty', 'DictProperty', 'SetProperty',
        'JsonObject', 'JsonArray',
    ]
else:
    __all__ = [
        b'IntegerProperty', b'FloatProperty', b'DecimalProperty',
        b'StringProperty', b'BooleanProperty',
        b'DateProperty', b'DateTimeProperty', b'TimeProperty',
        b'ObjectProperty', b'ListProperty', b'DictProperty', b'SetProperty',
        b'JsonObject', b'JsonArray',
    ]



from model.scenario import _ALIAS_Scenario
from model.model import _ALIAS_Model
from model.simulator import _ALIAS_Simulator
from model.dataframe_loader import _ALIAS_DataframeLoader
from config import config

if __name__ == "__main__":
    simulator = _ALIAS_Simulator(
        df_loader_cls=_ALIAS_DataframeLoader,
        config=config,
        scenario_cls=_ALIAS_Scenario,
        model_cls=_ALIAS_Model,
    )

    """
    Run the model
    """
    simulator.run()



# -*- coding:utf-8 -*-
# @Time: 2022/11/21 14:10
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: ws_protocol.py
import http
import logging
import os.path
from typing import Optional
import mimetypes
from websockets.datastructures import Headers
from websockets.legacy.server import WebSocketServerProtocol, HTTPResponse

from .actions import Action

logger = logging.getLogger(__name__)


def create_file_response(filepath: str):
    basename, ext = os.path.splitext(filepath)
    mime = mimetypes.types_map[ext]
    logger.info('ext-mime' + f'{mime}')
    with open(filepath, 'rb') as f:
        content = f.read()
    return (http.HTTPStatus.OK, [("Access-Control-Allow-Origin", "*"), ('Content-Type', mime)],
            content)


def create_failed_response(msg: str):
    return http.HTTPStatus.BAD_REQUEST, [("Access-Control-Allow-Origin", "*")], msg.encode('utf8')


class MelodieVisualizerProtocol(WebSocketServerProtocol):
    async def process_request(
            self, path: str, request_headers: Headers
    ) -> Optional[HTTPResponse]:
        """
        Handle requests in visualizer.

        :param path:
        :param request_headers:
        :return:
        """
        if path.startswith("/fs/"):
            filepath = path[4:]
            return create_file_response(filepath)
        elif path.startswith("/action/"):
            action = path[len('/action/'):]
            return Action.dispatch(action)
        ret = await super().process_request(path, request_headers)
        return ret



from Melodie import DataLoader


class _ALIAS_DataframeLoader(DataLoader):
    def register_scenario_dataframe(self):
        pass

    def register_static_dataframes(self):
        pass

    def generate_dataframe(self):
        pass



# -*- coding:utf-8 -*-
# @Time: 2022/12/13 21:50
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: ast_manipulator.py
import ast
import astunparse
from typing import Callable, List, Optional, Tuple


def load_func(file: str, func_name: str):
    with open(file, 'r', encoding='utf8', errors='replace') as f:
        root = ast.parse(f.read())
    for node in ast.walk(root):
        if isinstance(node, ast.FunctionDef) and node.name == func_name:
            return root, node
    raise Exception(f"Function {func_name} not in {file}!")


class ASTManipulator:
    def __init__(self):
        self.ast: Optional[ast.AST] = None


class Function:
    def __init__(self, name: str):
        self.name = name


class FuncDefManipulator(ASTManipulator):
    # ast: ast.FunctionDef
    # root:
    def __init__(self, file: str, func_name: str):
        super().__init__()
        self.ast: ast.FunctionDef
        self.root: ast.Module
        self.root, self.ast = load_func(file, func_name)
        self.functions: List[Function] = [Function("self.add_agent_property")]

    @property
    def body(self) -> List[ast.stmt]:
        return self.ast.body

    # @property
    def get_records(self, func_name: str):
        records = []
        for node in ast.walk(self.ast):
            if self.is_valid_ast_node(func_name, node):
                records.append(self.parse_valid_node(node))
        return records

    def add_record(self, func_name: str, record: Tuple[str, str]):
        self.body.append(
            ast.Expr(
                value=ast.Call(
                    func=ast.parse(func_name),
                    args=[ast.Constant(value=s) for s in record],
                    keywords=[]
                )
            )
        )

    def remove_record(self, func_name: str, record: Tuple[str, str]):
        i = -1
        for i, item in enumerate(self.body):
            if isinstance(item, ast.Expr):
                node = item.value
                if self.is_valid_ast_node(func_name, node) and self.parse_valid_node(node) == record:
                    break
            else:
                continue

        if i >= 0:
            self.body.pop(i)

    @staticmethod
    def is_valid_ast_node(func_name: str, node: ast.AST):
        if isinstance(node, ast.Expr) and isinstance(node.value, ast.Call):
            node = node.value
        if isinstance(node, (ast.Call,)):
            fname = astunparse.unparse(node.func).strip()
            if fname == func_name:
                return True

        return False

    def parse_valid_node(self, node: ast.Call):
        args = node.args
        value = []
        for arg in args:
            assert isinstance(arg, ast.Constant)
            value.append(arg.value)
        return tuple(value)



import logging
import time
from typing import List, TYPE_CHECKING, Dict, Tuple, Any, Optional, Type

import pandas as pd

from MelodieInfra import DBConn, MelodieExceptions

from Melodie.global_configs import MelodieGlobalConfig


logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from Melodie import Model, Scenario, BaseAgentContainer


class PropertyToCollect:
    def __init__(self, property_name: str, as_type: Type):
        self.property_name = property_name
        self.as_type = as_type


class DataCollector:
    """
    Data Collector collects data in the model.
    
    At the beginning of simulation, the ``DataCollector`` creates as the model creates;
    
    User could customize which property of agents or environment should be collected.
    
    Before the model running finished, the DataCollector dumps data to dataframe, and save to
    database.
    """

    def __init__(self, target="sqlite"):
        if target not in {"sqlite", None}:
            MelodieExceptions.Data.InvalidDatabaseType(target, {"sqlite"})

        self.target = target
        self.model: Optional[Model] = None
        self.scenario: Optional["Scenario"] = None
        self._agent_properties_to_collect: Dict[str, List[PropertyToCollect]] = {}
        self._environment_properties_to_collect: List[PropertyToCollect] = []

        self.agent_properties_df = pd.DataFrame()
        self.environment_properties_df = pd.DataFrame()

        self.agent_properties_dict: Dict[str, List[Any]] = {}
        self.environment_properties_list = []

        self._time_elapsed = 0

    def setup(self):
        """
        Setup method, be sure to inherit it.

        :return:
        """
        pass

    def _setup(self):
        self.setup()
        self._time_elapsed = 0

    def time_elapsed(self):
        """
        Get the time spent in data collection.

        :return: Elapsed time, a ``float`` value.
        """
        return self._time_elapsed

    def add_agent_property(
        self, container_name: str, property_name: str, as_type: Type = None
    ):
        """
        This method adds a property of agents in an agent_container to the data collector.
        
        The type which the data will be represented in the database can also be determined by ``as_type``.

        :param container_name: Container name, also a property name on model.
        :param property_name: Property name on the agent.
        :param as_type: Data type.
        :return:
        """
        if not hasattr(self.model, container_name):
            raise AttributeError(f"Model has no agent container '{container_name}'")
        if container_name not in self._agent_properties_to_collect.keys():
            self._agent_properties_to_collect[container_name] = []
        self._agent_properties_to_collect[container_name].append(
            PropertyToCollect(property_name, as_type)
        )

    def add_environment_property(self, property_name: str, as_type: Type = None):
        """
        This method tells the data collector which property of environment should be collected.

        :param property_name: Environment properties
        :param as_type: Data type.
        :return:
        """
        self._environment_properties_to_collect.append(
            PropertyToCollect(property_name, as_type)
        )

    def env_property_names(self) -> List[str]:
        """
        Get the environment property names to collect

        :return: List of environment property names
        """
        return [prop.property_name for prop in self._environment_properties_to_collect]

    def agent_property_names(self) -> Dict[str, List[str]]:
        """
        Get the agent property names to collect

        :return: A ``dict``,  ``<agent_container_name --> agent list properties to gather>[]``
        """
        return {
            container_name: [prop.property_name for prop in props]
            for container_name, props in self._agent_properties_to_collect.items()
        }

    def agent_containers(self) -> List[Tuple[str, "BaseAgentContainer"]]:
        """
        Get all agent containers with attributes to collect in the model.

        :return: A list of tuples, ``<agent_container_name, agent_container_object>[]``
        """
        containers = []
        for container_name in self._agent_properties_to_collect.keys():
            containers.append((container_name, getattr(self.model, container_name)))
        return containers

    def collect_agent_properties(self, period: int, id_run: int, id_scenario: int):
        """
        Collect agent properties.

        :param period: Current simulation step
        :param id_run: Current simulation ``run_id``
        :param id_scenario: Current scenario id
        :return: None
        """
        agent_containers = self.agent_containers()
        agent_property_names = self.agent_property_names()
        for container_name, container in agent_containers:
            agent_prop_list = container.to_list(agent_property_names[container_name])
            length = len(agent_prop_list)
            props_list = []
            for i in range(length):
                agent_props_dict = agent_prop_list[i]
                tmp_dic = {
                    "id_scenario": id_scenario,
                    "id_run": id_run,
                    "period": period,
                    "id": agent_props_dict.pop("id"),
                }
                tmp_dic.update(agent_props_dict)
                props_list.append(tmp_dic)

            if container_name not in self.agent_properties_dict:
                self.agent_properties_dict[container_name] = []
            self.agent_properties_dict[container_name].extend(props_list)

    @property
    def status(self) -> bool:
        """
        If data collector is enabled.

        ``DataCollector`` is only enabled in the ``Simulator``, because ``Trainer`` and ``Calibrator`` are only concerned over
        the properties at the end of the model-running, so recording middle status in ``Trainer`` or ``Calibrator`` is a waste of time and space.

        :return: bool.
        """
        from .simulator import Simulator

        operator = self.model.scenario.manager
        return isinstance(operator, Simulator)

    def collect(self, period: int) -> None:
        """
        The main function to collect data.

        :param period: Current simulation step.
        :return: None
        """
        if not self.status:
            return
        t0 = time.time()

        env_dic = {
            "id_scenario": self.model.scenario.id,
            "id_run": self.model.run_id_in_scenario,
            "period": period,
        }
        env_dic.update(self.model.environment.to_dict(self.env_property_names()))

        self.environment_properties_list.append(env_dic)

        self.collect_agent_properties(
            period, self.model.run_id_in_scenario, self.model.scenario.id
        )
        t1 = time.time()
        self._time_elapsed += t1 - t0

    @property
    def db(self):
        """
        Create a database connection

        :return: ``melodie.DB`` object.
        """
        return self.model.create_db_conn()

    @staticmethod
    def calc_time(method):
        """
        Works as a decorator.

        If you would like to define a custom data-collect method, please use ``DataCollector.calc_time`` as a decorator.
        """

        def wrapper(obj: DataCollector, *args, **kwargs):
            t0 = time.time()
            ret = method(obj, *args, **kwargs)
            t1 = time.time()
            obj._time_elapsed += t1 - t0
            return ret

        return wrapper

    def save(self):
        """
        Save the collected data into database.

        :return: None
        """
        if not self.status:
            return
        t0 = time.time()
        write_db_time = 0
        connection = self.model.create_db_conn()
        environment_properties_df = pd.DataFrame(self.environment_properties_list)
        _t = time.time()
        connection.write_dataframe(
            DBConn.ENVIRONMENT_RESULT_TABLE, environment_properties_df, {}
        )
        write_db_time += time.time() - _t

        for container_name in self.agent_properties_dict.keys():
            agent_properties_df = pd.DataFrame(
                self.agent_properties_dict[container_name]
            )
            _t = time.time()
            connection.write_dataframe(
                container_name + "_result", agent_properties_df, {}
            )
            write_db_time += time.time() - _t

        t1 = time.time()
        collect_time = self._time_elapsed
        self._time_elapsed += t1 - t0
        logger.debug(
            f"datacollector took {MelodieGlobalConfig.Logger.round_elapsed_time(t1 - t0)}s to format dataframe and write it to data.\n"
            f"    {MelodieGlobalConfig.Logger.round_elapsed_time(write_db_time)} for writing into database, and "
            f"{MelodieGlobalConfig.Logger.round_elapsed_time(collect_time)} for collect data."
        )



"""
This data stores the run function for model running, storing global variables and other services.
"""
import abc
import logging
import os
import threading
import time
from multiprocessing import Pool
from typing import Optional, List, Tuple, Type

import numpy as np
import pandas as pd

from MelodieInfra import create_db_conn, get_sqlite_filename, DBConn, Config, MelodieExceptions, show_prettified_warning

from .global_configs import MelodieGlobalConfig
from .data_loader import DataLoader
from .model import Model
from .scenario_manager import Scenario
from .visualizer import BaseVisualizer, MelodieModelReset

logger = logging.getLogger(__name__)


class BaseModellingManager(abc.ABC):
    """
    Base class of ``Simulator``/``Trainer``/``Calibrator``.
    """

    def __init__(
            self,
            config: Config,
            scenario_cls: Type["Scenario"],
            model_cls: Type["Model"],
            data_loader_cls: Type[DataLoader] = None,
    ):
        self.config: Optional[Config] = config
        self.scenario_cls = scenario_cls
        self.model_cls = model_cls

        self.scenarios: Optional[List["Scenario"]] = None
        self.df_loader_cls: Optional[Type[DataLoader]] = data_loader_cls
        self.data_loader: Optional[DataLoader] = None
        if data_loader_cls is not None:
            assert issubclass(data_loader_cls, DataLoader), data_loader_cls

    def get_dataframe(self, table_name) -> pd.DataFrame:
        """
        Get a static ``DataFrame`` from the data loader.

        :param table_name: Name of dataframe.
        :return:
        """
        if self.data_loader is None:
            raise MelodieExceptions.Data.NoDataframeLoaderDefined()
        if table_name not in self.data_loader.registered_dataframes:
            raise MelodieExceptions.Data.StaticTableNotRegistered(
                table_name, list(self.data_loader.registered_dataframes.keys())
            )

        return self.data_loader.registered_dataframes[table_name]

    def get_matrix(self, matrix_name) -> np.ndarray:
        """
        Get a matrix from data loader.

        :param matrix_name: Name of matrix.
        :return:
        """
        if self.data_loader is None:
            raise MelodieExceptions.Data.NoDataframeLoaderDefined()
        if matrix_name not in self.data_loader.registered_matrices:
            raise MelodieExceptions.Data.StaticTableNotRegistered(
                matrix_name, list(self.data_loader.registered_matrices.keys())
            )

        return self.data_loader.registered_matrices[matrix_name]

    def subworker_prerun(self):
        """
        If run as sub-worker, run this function to avoid deleting the existing tables in database.
        """
        self.data_loader: DataLoader = self.df_loader_cls(
            self, self.config, self.scenario_cls, as_sub_worker=True
        )

        self.scenarios = self.generate_scenarios()

    def pre_run(self, clear_db=True):
        """
        `pre_run` means this function should be executed before ``run``, to initialize the scenario
        parameters.

        :param clear_db: If ``True``, this method will clear database.
        :return:
        """
        if clear_db:
            create_db_conn(self.config).clear_database()
        if self.df_loader_cls is not None:
            self.data_loader: DataLoader = self.df_loader_cls(
                self, self.config, self.scenario_cls
            )

        self.scenarios = self.generate_scenarios()
        if self.scenarios is None or len(self.scenarios) == 0:
            MelodieExceptions.Scenario.NoValidScenarioGenerated(self.scenarios)

    @abc.abstractmethod
    def generate_scenarios(self):
        """
        Abstract method for generation of scenarios.
        """
        pass


class Simulator(BaseModellingManager):
    def __init__(
            self,
            config: Config,
            scenario_cls: "Type[Scenario]",
            model_cls: "Type[Model]",
            data_loader_cls: "Type[DataLoader]" = None,
            visualizer_cls: "type[BaseVisualizer]" = None,
    ):
        super(Simulator, self).__init__(
            config=config,
            scenario_cls=scenario_cls,
            model_cls=model_cls,
            data_loader_cls=data_loader_cls,
        )
        self.server_thread: Optional[threading.Thread] = None
        self.visualizer_cls = visualizer_cls
        self.visualizer = None

    def _init_visualizer(self):
        self.visualizer: Optional[BaseVisualizer] = (
            None if self.visualizer_cls is None else self.visualizer_cls(self.config, self)
        )

    def generate_scenarios(self) -> List["Scenario"]:
        """
        Generate scenarios from the dataframe_loader

        :return: A list of generated scenarios.
        """
        if self.data_loader is None:
            raise MelodieExceptions.Data.NoDataframeLoaderDefined()
        return self.data_loader.generate_scenarios("simulator")

    def run_model(
            self, config, scenario, id_run, model_class: Type["Model"], visualizer=None
    ):
        """
        Run a model once.
        """
        logger.info(
            f"Simulation started - id_scenario = {scenario.id}, id_run = {id_run}"
        )
        t0 = time.time()

        model: Model = model_class(
            config, scenario, run_id_in_scenario=id_run, visualizer=visualizer
        )
        if visualizer is not None:
            visualizer.set_model(model)
            if not self.visualizer.params_manager._initialized:
                self.visualizer.params_manager.write_obj_attrs_to_params_list(scenario,
                                                                              self.visualizer.params_manager.params)
                self.visualizer.params_manager._initialized = True
            else:
                self.visualizer.params_manager.modify_scenario(scenario)
            # with open('test.json', 'w') as f:
            #     json.dump({
            #         'model': self.visualizer.params_manager.to_json(),
            #         'params-values': self.visualizer.params_manager.to_value_json()
            #     }, f, indent=4)
            # with open('123.json') as f:
            #     self.visualizer.params_manager.from_json(json.load(f))
            # with open('out.json', 'w') as f:
            #     json.dump({
            #         'model': self.visualizer.params_manager.to_json(),
            #         'params-values': self.visualizer.params_manager.to_value_json()
            #     }, f, indent=4)
            # self.visualizer.params_manager.modify_scenario(scenario)
            # print('aaa', self.visualizer.params_manager.to_value_json()[0])
            visualizer.start()
        else:
            model._setup()

        t1 = time.time()
        model.run()
        if visualizer is not None:
            model.visualizer.finish()
        t2 = time.time()

        model_setup_time = t1 - t0
        model_run_time = t2 - t1
        if model.data_collector is not None:
            data_collect_time = model.data_collector.time_elapsed()
        else:
            data_collect_time = 0.0
        model_run_time -= data_collect_time
        info = (
            f"Simulation completed - id_scenario = {scenario.id}, id_run = {id_run}\n"
            f"    model-setup   \t {MelodieGlobalConfig.Logger.round_elapsed_time(model_setup_time)}s\n"
            f"    model-run     \t {MelodieGlobalConfig.Logger.round_elapsed_time(model_run_time)}s\n"
            f"    data-collect  \t {MelodieGlobalConfig.Logger.round_elapsed_time(data_collect_time)}s\n"
        )
        logger.info(info)

    def setup(self):
        """
        Setup method of simulator, usually need not to inherit.

        :return: None
        """
        pass

    def run(self):
        """
        Main function for running model.

        :return: None
        """
        t0 = time.time()
        BaseVisualizer.enabled = False
        self.setup()
        if self.visualizer_cls is not None:
            show_prettified_warning(
                "You are using `Simulator.run()` method to run model, but you have also defined visualizer.\n"
                "If you would like to visualize this model, please use `Simulator.run_visual()` "
                "instead of `Simulator.run()` "
            )

        self.pre_run()
        t1 = time.time()
        for scenario_index, scenario in enumerate(self.scenarios):
            for id_run in range(scenario.run_num):
                self.run_model(
                    self.config, scenario, id_run, self.model_cls, visualizer=None
                )

        t2 = time.time()
        logger.info(
            f"Simulator completed, time elapsed {MelodieGlobalConfig.Logger.round_elapsed_time(t2 - t0)}s."
        )

    def run_visual(self):
        """
        Main function for running model with visualization.

        :return: None
        """
        t0 = time.time()
        self._init_visualizer()
        self.setup()

        self.pre_run()
        if self.visualizer is not None:
            self.visualizer.setup()

        t1 = time.time()
        logger.info(f"Simulator start up cost: {t1 - t0}s")
        while True:
            logger.info(
                f"Visualizer interactive paramerters for this scenario are: {self.visualizer.scenario_param}"
            )
            # create_db_conn(self.config).clear_database()
            # create_db_conn()
            fn = get_sqlite_filename(self.config)
            if os.path.exists(fn):
                os.remove(fn)
            self.data_loader: "Optional[DataLoader]" = None
            self.scenarios = None
            DBConn.table_dtypes = {}
            self.pre_run()
            scenario = self.scenarios[0].copy()
            scenario.manager = self

            # for k, v in self.visualizer.scenario_param.items():
            #     scenario.__setattr__(k, v)
            logger.info(f"Scenario parameters: {scenario.to_dict()}")
            try:
                self.visualizer.current_scenario = scenario  # set studio scenario.
                self.run_model(
                    self.config, scenario, 0, self.model_cls, visualizer=self.visualizer
                )
            except MelodieModelReset:

                self.visualizer.reset()
                logger.info("Model reset.")

    def run_parallel(self, cores: int = 1):
        """
        Parallelized running through a series of scenarios.

        Melodie does not start subprocesses directly. For the first shot, which means running one of the runs out of
        one scenario, it will be run by the main-thread to verify the model and initialize the database.
        After the first shot, subprocesses will be created as many as the value of parameter `cores`.

        :param cores: How many subprocesses will be created in the parallel simulation.

          - It is suggested that this parameter should be NO MORE THAN the **physical cores** of your computer.
          - Beside 'cores', You may found that your cpu has one more metric: threads, which means your CPU supports
            hyper-threading. If so, use 'physical cores', not 'threads' as the upper limit.
          - For example, an Intel® I5-8250U has 4 physical cores and 8 threads. If you use a computer equipped with
            this CPU, this parameter cannot be larger than 4.
          - In short, hyper-threading only improves performance in io intensive programs. As Melodie was computation
            intensive, if there are more subprocess than physical cores, subprocesses will fight for CPU, costing
            a lot of extra time.

        Sqlite itself was thread-safe for writing. However, pandas tries to create the table if table was not exist, which
        might trigger conflict condition.

        For example:
            p1, p2 stands for process 1 and 2 request writing to a table `test`;
            db stands for database;
            1. p1 --> db <found no table named 'test'>
            2. p2 --> db <found no table named 'test'>
            3. p1 --> db <request the db to create a table 'test'>
            4. p2 --> db <request the db to create a table 'test'>
            5. db --> p1 <created table named 'test'>
            6. db --> p2 <table 'test' already exists!> ERROR!

        For multiple-cores, if running modelthis might happen very frequently. To avoid this, Melodie makes the first
        shot, which means running one of the runs out of one scenario, by main-process. Only when first shot completes
        will the subprocesses be launched.


        :return: None
        """
        t0 = time.time()
        self.pre_run()

        t1 = time.time()
        pool = Pool(cores)

        parameters: List[Tuple] = []
        for scenario_index, scenario in enumerate(self.scenarios):
            for id_run in range(scenario.run_num):
                params = (
                    self.config,
                    scenario,
                    id_run,
                    self.model_cls,
                )
                parameters.append(params)

        logger.info(f"Melodie will run totally {len(parameters)} times!.")
        logger.info(
            "Running the first session with only one core to verify this model..."
        )
        self.run_model(*parameters.pop())
        logger.info(
            f"Verification finished, now using {cores} cores for parallel computing!"
        )
        pool.starmap(self.run_model, parameters)

        pool.close()
        pool.join()
        t2 = time.time()
        logger.info(
            f"Melodie completed all runs, time elapsed totally {t2 - t0}s, and {t2 - t1}s for running."
        )



import copy
import logging
from typing import (
    Dict,
    Tuple,
    Callable,
    Union,
    List,
    Any,
    Optional,
    Iterator,
    Type,
    cast,
)

import pandas as pd
from sko import GA

from MelodieInfra import Config, MelodieExceptions, create_db_conn

from .algorithms import AlgorithmParameters
from .algorithms.ga import MelodieGA
from .boost.agent_list import AgentList
from .boost.basics import Agent
from .data_loader import DataLoader
from .model import Model
from .scenario_manager import Scenario
from .simulator import BaseModellingManager
from .utils.parallel_manager import ParallelManager

logger = logging.getLogger(__name__)
pool = None  # for *nix
th_on_thread = None  # for windows


class GATrainerParams(AlgorithmParameters):
    def __init__(
            self,
            id: int,
            path_num: int,
            generation_num: int,
            strategy_population: int,
            mutation_prob: float,
            strategy_param_code_length: int,
            **kw,
    ):
        super().__init__(id, path_num)
        self.generation_num = generation_num
        self.strategy_population = strategy_population
        self.mutation_prob = mutation_prob
        self.strategy_param_code_length = strategy_param_code_length
        self.parse_params(kw)

    @staticmethod
    def from_dataframe_record(
            record: Dict[str, Union[int, float]]
    ) -> "GATrainerParams":
        s = GATrainerParams(
            record["id"],
            record["path_num"],
            record["generation_num"],
            record["strategy_population"],
            record["mutation_prob"],
            record["strategy_param_code_length"],
        )
        s.parse_params(record)
        return s

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.__dict__}>"

    def __hash__(self):
        return hash(
            (
                self.id,
                self.path_num,
                self.generation_num,
                self.strategy_population,
                self.mutation_prob,
                self.strategy_param_code_length,
            )
        )


class TrainerAlgorithmMeta:
    """
    Record the current scenario, params scenario, path and generation
    of trainer.

    """

    def __init__(self):
        self._freeze = False
        self.id_trainer_scenario = 0
        self.id_trainer_params_scenario = 1
        self.id_path = 0
        self.id_generation = 0

    def to_dict(self, public_only=False):

        if public_only:
            return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        return copy.copy(self.__dict__)

    def __repr__(self):
        return f"<{self.to_dict()}>"

    def __setattr__(self, key, value):
        if (not hasattr(self, "_freeze")) or (not self._freeze):
            super().__setattr__(key, value)
        else:
            if key in self.__dict__:
                super(TrainerAlgorithmMeta, self).__setattr__(key, value)
            else:
                raise MelodieExceptions.General.NoAttributeError(self, key)


class GATrainerAlgorithmMeta(TrainerAlgorithmMeta):
    def __init__(self):
        super().__init__()
        self.id_chromosome = 0
        self._freeze = True


class TargetFcnCache:
    """
    The cache for the pre-computed values.

    Dict:  {(generation_id,
             id_chromosome):
                           {(agent_id,
                             container_name):
                                            function_value
                           }
          }
    """

    def __init__(self):
        self.target_fcn_record: Dict[Tuple[int, int], Dict] = {}
        self.current_generation = -1
        self.current_id_chromosome = -1

    def lookup_agent_target_value(
            self, agent_id: int, container_name: str, generation: int, id_chromosome: int
    ):
        return self.target_fcn_record[(generation, id_chromosome)][
            (agent_id, container_name)
        ]

    def set_agent_target_value(
            self,
            agent_id: int,
            container_name: str,
            value: float,
            generation: int,
            id_chromosome: int,
    ):
        # self.current_target_fcn_value[(agent_id, container_name)] = value
        if (generation, id_chromosome) not in self.target_fcn_record:
            self.target_fcn_record[(generation, id_chromosome)] = {}
        self.target_fcn_record[(generation, id_chromosome)][
            (agent_id, container_name)
        ] = value

    def best_value(
            self, chromosome_num: int, generation: int, agent_id: int, agent_category: int
    ):
        values = [
            self.target_fcn_record[(generation, id_chromosome)][
                (agent_id, agent_category)
            ]
            for id_chromosome in range(chromosome_num)
        ]
        return min(values)


class GATrainerAlgorithm:
    """
    参数：一个tuple
    每次会跑20条染色体，然后将参数缓存起来。
    目标函数从TargetFcnCache中查询。
    """

    def __init__(
            self, params: GATrainerParams, manager: "Trainer" = None, processors=1
    ):
        global pool, th_on_thread
        self.manager = manager
        self.params = params
        self.chromosomes = 20
        self.algorithms_dict: Dict[Tuple[int, str], Union[GA]] = {}

        self.target_fcn_cache = TargetFcnCache()
        self.agent_container_getters: Dict[str, Callable[[Model], AgentList]] = {}
        self.agent_ids: Dict[str, List[int]] = {}  # {category : [agent_id]}
        self.agent_params_defined: Dict[
            str, List[str]
        ] = {}  # category: [param_name1, param_name2...]
        self.recorded_agent_properties: Dict[
            str, List[str]
        ] = {}  # category: [prop1, prop2...]
        self.recorded_env_properties: List[str] = []  # category: [prop1, prop2...]

        self._chromosome_counter = 0
        self._current_generation = 0
        self.processors = processors

        # for i in range(processors):
        d = {
            "model": (
                self.manager.model_cls.__name__,
                self.manager.model_cls.__module__,
            ),
            "scenario": (
                self.manager.scenario_cls.__name__,
                self.manager.scenario_cls.__module__,
            ),
            "trainer": (
                self.manager.__class__.__name__,
                self.manager.__class__.__module__,
            ),
            "data_loader": (
                self.manager.df_loader_cls.__name__,
                self.manager.df_loader_cls.__module__,
            ),
        }
        self.parallel_manager = ParallelManager(
            self.processors, configs=(d, self.manager.config.to_dict())
        )
        self.parallel_manager.run("trainer")

    def stop(self):
        self.parallel_manager.close()

    def setup_agent_locations(
            self,
            container_name: str,
            param_names: List[str],
            recorded_properties: List[str],
            agent_id_list: List[int],
    ):
        assert container_name not in self.agent_container_getters
        self.agent_container_getters[container_name] = lambda model: getattr(
            model, container_name
        )
        lb, ub = self.params.bounds(param_names)
        for agent_id in agent_id_list:
            self.algorithms_dict[(agent_id, container_name)] = MelodieGA(
                func=cast(
                    "function", self.generate_target_function(agent_id, container_name)
                ),
                n_dim=len(param_names),
                size_pop=self.params.strategy_population,
                max_iter=self.params.generation_num,
                prob_mut=self.params.mutation_prob,
                lb=lb,
                ub=ub,
                precision=1e-5,
            )
        self.agent_params_defined[container_name] = param_names
        self.recorded_agent_properties[container_name] = recorded_properties
        self.agent_ids[container_name] = agent_id_list

    def get_agent_params(self, id_chromosome: int):
        """
        Pass parameters from the chromosome to the agent container.

        :param id_chromosome:
        :return:
        """
        params: Dict[str, List[Dict[str, Any]]] = {
            category: [] for category in self.agent_ids.keys()
        }
        # {category : [{id: 0, param1: 1, param2: 2, ...}]}
        for key, algorithm in self.algorithms_dict.items():
            chromosome_value = algorithm.chrom2x(algorithm.Chrom)[id_chromosome]
            agent_id, agent_category = key
            d = {"id": agent_id}
            for i, param_name in enumerate(self.agent_params_defined[agent_category]):
                d[param_name] = chromosome_value[i]
            params[agent_category].append(d)
        return params

    def target_function_to_cache(
            self,
            agent_target_function_values: Dict[str, List[Dict[str, Any]]],
            generation: int,
            id_chromosome: int,
    ):
        """
        Extract the value of target functions from Model, and write them into cache.

        :return:
        """
        for (
                container_category,
                container_getter,
        ) in self.agent_container_getters.items():
            agent_props_list = agent_target_function_values[container_category]
            for agent_props in agent_props_list:
                self.target_fcn_cache.set_agent_target_value(
                    agent_props["agent_id"],
                    container_category,
                    agent_props["target_function_value"],
                    generation,
                    id_chromosome,
                )

    def generate_target_function(
            self, agent_id: int, container_name: str
    ) -> Callable[[], float]:
        def f(*args):
            self._chromosome_counter += 1
            value = self.target_fcn_cache.lookup_agent_target_value(
                agent_id,
                container_name,
                self._current_generation,
                self._chromosome_counter,
            )
            return value

        return f

    def record_agent_properties(
            self,
            agent_data: Dict[str, List[Dict[str, Any]]],
            env_data: Dict[str, Any],
            meta: GATrainerAlgorithmMeta,
    ):
        """
        Record the property of each agent in the current chromosome.

        :param agent_data: {<agent_container_name>: [{id: 0, prop1: xxx, prop2: xxx, ...}]}
        :param env_data: {env_prop1: xxx, env_prop2: yyy, ...}
        :param meta:
        :return:
        """
        agent_records = {}
        env_record = {}
        meta_dict = meta.to_dict(public_only=True)

        for (
                container_name,
                agent_container_getter,
        ) in self.agent_container_getters.items():
            agent_records[container_name] = []
            data = agent_data[container_name]
            for agent_container_data in data:
                d = {}
                d.update(meta_dict)
                d["agent_id"] = agent_container_data["agent_id"]
                d.update(agent_container_data)
                d.pop("target_function_value")
                agent_records[container_name].append(d)
            create_db_conn(self.manager.config).write_dataframe(
                f"{container_name}_trainer_result",
                pd.DataFrame(agent_records[container_name]),
                if_exists="append",
            )
        env_record.update(meta_dict)
        env_record.update(env_data)

        create_db_conn(self.manager.config).write_dataframe(
            "environment_trainer_result", pd.DataFrame([env_record]), if_exists="append"
        )

        return agent_records, env_record

    def calc_cov_df(
            self,
            agent_container_df_dict: Dict[str, pd.DataFrame],
            env_df: pd.DataFrame,
            meta,
    ):
        """
        Calculate the coefficient of variation
        :param agent_container_df_dict:
        :param env_df:
        :param meta:
        :return:
        """

        pd.set_option("max_colwidth", 500)
        pd.set_option("display.max_columns", None)
        pd.set_option("display.max_rows", None)
        meta_dict = meta.to_dict(public_only=True)
        meta_dict.pop("id_chromosome")
        for container_name in self.agent_ids.keys():
            df = agent_container_df_dict[container_name]
            container_agent_record_list = []
            for agent_id in self.agent_ids[container_name]:
                agent_data = df.loc[df["agent_id"] == agent_id]
                cov_records = {}
                cov_records.update(meta_dict)
                cov_records["agent_id"] = agent_id
                for prop_name in self.recorded_agent_properties[container_name] + [
                    "utility"
                ]:
                    p: pd.Series = agent_data[prop_name]
                    mean = p.mean()
                    cov = p.std() / p.mean()
                    cov_records.update(
                        {prop_name + "_mean": mean, prop_name + "_cov": cov}
                    )
                container_agent_record_list.append(cov_records)
            create_db_conn(self.manager.config).write_dataframe(
                f"{container_name}_trainer_result_cov",
                pd.DataFrame(container_agent_record_list),
                if_exists="append",
            )
        env_record = {}
        env_record.update(meta_dict)
        for prop_name in self.recorded_env_properties:
            mean = env_df[prop_name].mean()
            cov = env_df[prop_name].std() / env_df[prop_name].mean()
            env_record.update({prop_name + "_mean": mean, prop_name + "_cov": cov})
        create_db_conn(self.manager.config).write_dataframe(
            "environment_trainer_result_cov", pd.DataFrame([env_record]), if_exists="append"
        )

    def pre_check(self, meta):
        """
        Check at the beginning of run()
        :return:
        """
        logger.info(f"""Algorithm will run with:
        Meta value: {meta}
        Recording environment parameters: {self.recorded_env_properties}
        Recording Agent agent_lists: {self.recorded_agent_properties}\n""")
        # print("Algorithm will run with:")
        # print("    Meta value", meta)
        # print("    Recording environment parameters: ", self.recorded_env_properties)
        # print("    Recording Agent agent_lists:", self.agent_container_getters)

    def run(self, scenario: Scenario, meta: Union[GATrainerAlgorithmMeta]):
        self.pre_check(meta)

        for i in range(self.params.generation_num):
            self._current_generation = i
            meta.id_generation = i
            logger.info(
                f"======================="
                f"Path {meta.id_path} Generation {i + 1}/{self.params.generation_num}"
                f"======================="
            )

            for id_chromosome in range(self.params.strategy_population):
                params = self.get_agent_params(id_chromosome)
                self.parallel_manager.put_task(
                    (id_chromosome, scenario.to_json(), params)
                )
                # params_queue.put(
                #     json.dumps()
                # )

            agent_records_collector: Dict[str, List[Dict[str, Any]]] = {
                container_name: [] for container_name in self.agent_ids.keys()
            }
            env_records_list: List[Dict[str, Any]] = []

            for _id_chromosome in range(self.params.strategy_population):
                # v = result_queue.get()

                (
                    chrom,
                    agents_data,
                    env_data,
                ) = (
                    self.parallel_manager.get_result()
                )  # cloudpickle.loads(base64.b64decode(v))
                meta.id_chromosome = chrom
                agent_records, env_record = self.record_agent_properties(
                    agents_data, env_data, meta
                )
                for container_name, records in agent_records.items():
                    agent_records_collector[container_name] += records
                env_records_list.append(env_record)
                self.target_function_to_cache(agents_data, i, chrom)

            self.calc_cov_df(
                {k: pd.DataFrame(v) for k, v in agent_records_collector.items()},
                pd.DataFrame(env_records_list),
                meta,
            )

            for key, algorithm in self.algorithms_dict.items():
                self._chromosome_counter = -1
                algorithm.run(1)


class RelatedAgentContainerModel:
    def __init__(
            self,
            container_name: str,
            used_properties: List[str],
            recorded_properties: List[str],
            agent_ids: Optional[Callable[[Scenario], List[int]]] = None,
    ):
        self.container_name = container_name
        self.used_properties = used_properties
        self.recorded_properties = recorded_properties

        self.agent_ids: Optional[Callable[[Scenario], List[int]]] = agent_ids


class AgentContainerManager:
    """
    The manager class is designed to organize the metadata related to agent agent_lists in the trainer.

    """

    def __init__(self):
        self.agent_containers: List[RelatedAgentContainerModel] = []

    def add_container(
            self,
            container_name: str,
            used_properties: List[str],
            agent_ids: Optional[Callable[[Scenario], List[int]]],
    ):
        """
        Add a container used in trainer.

        :param container_name:
        :param used_properties:
        :param recorded_properties:
        :param agent_ids:
        :return:
        """
        self.agent_containers.append(
            RelatedAgentContainerModel(container_name, used_properties, [], agent_ids)
        )

    def get_agent_container(
            self, agent_container_name: str
    ) -> RelatedAgentContainerModel:
        for agent_container_model in self.agent_containers:
            if agent_container_model.container_name == agent_container_name:
                return agent_container_model
        raise KeyError(agent_container_name)


class Trainer(BaseModellingManager):
    """
    Individually train agents' parameters
    """

    def __init__(
            self,
            config: "Config",
            scenario_cls: "Optional[Type[Scenario]]",
            model_cls: "Optional[Type[Model]]",
            data_loader_cls: "Optional[Type[DataLoader]]",
            processors: int = 1,
    ):
        super().__init__(
            config=config,
            scenario_cls=scenario_cls,
            model_cls=model_cls,
            data_loader_cls=data_loader_cls,
        )
        self.container_manager = AgentContainerManager()

        self.agent_result_properties: Dict[str, List[str]] = {}

        self.environment_properties: List[str] = []

        self.algorithm_type: str = "ga"
        self.algorithm: Optional[GATrainerAlgorithm] = None
        self.algorithm_instance: Iterator[List[float]] = {}

        self.save_agent_trainer_result = False
        self.save_env_trainer_result = True

        self.model: Optional[Model] = None

        self.agent_result = []
        self.current_algorithm_meta = None
        self.processors = processors

    def add_agent_training_property(
            self,
            agent_list_name: str,
            training_attributes: List[str],
            agent_ids: Callable[[Scenario], List[int]]
    ):
        """
        Add a container into the trainer.

        :param agent_list_name: The name of agent container.
        :param training_attributes: The properties used in training.
        :param agent_ids: The agent with id contained in `agent_ids` will be trained.
        :return: None
        """
        self.container_manager.add_container(agent_list_name, training_attributes, agent_ids)

    def setup(self):
        """
        Setup method, be sure to inherit this method in custom trainer class.
        """
        pass

    def get_trainer_scenario_cls(self):
        """
        Get the class of trainer scenario.

        :return: Trainer parameters
        """
        assert self.algorithm_type in {"ga"}

        trainer_scenario_cls: Optional[Union[Type[GATrainerParams]]] = None
        if self.algorithm_type == "ga":
            trainer_scenario_cls = GATrainerParams
        assert trainer_scenario_cls is not None
        return trainer_scenario_cls

    def collect_data(self):
        """
        Set the agent and environment properties to be collected.

        :return:
        """
        pass

    def run(self):
        """
        The main method for Trainer.

        :return:
        """
        self.setup()
        self.collect_data()
        self.pre_run()

        trainer_scenario_cls = self.get_trainer_scenario_cls()
        self.current_algorithm_meta = GATrainerAlgorithmMeta()
        for scenario in self.scenarios:
            self.current_algorithm_meta.id_trainer_scenario = scenario.id
            for trainer_params in self.generate_trainer_params_list(
                    trainer_scenario_cls
            ):
                self.current_algorithm_meta.id_trainer_params_scenario = trainer_params.id
                for id_path in range(trainer_params.path_num):
                    self.current_algorithm_meta.id_path = id_path
                    logger.info(
                        f"id_trainer_scenario = {scenario.id}, id_path = {id_path}"
                    )
                    self.run_once_new(scenario, trainer_params)

    def run_once_new(self, scenario: Scenario, trainer_params: Union[GATrainerParams]):
        """
        Run for one training path

        :param scenario: The scenario to run
        :param trainer_params: calibration parameters.
        :return: None
        """

        self.algorithm = GATrainerAlgorithm(trainer_params, self, self.processors)
        self.algorithm.recorded_env_properties = self.environment_properties
        for agent_container in self.container_manager.agent_containers:
            self.algorithm.setup_agent_locations(
                agent_container.container_name,
                agent_container.used_properties,
                agent_container.recorded_properties,
                agent_container.agent_ids(scenario),
            )

        self.algorithm.run(scenario, self.current_algorithm_meta)
        self.algorithm.stop()

    def utility(self, agent: Agent) -> float:
        """
        The utility is to be maximized.
        be sure to inherit inside the custom trainer class.

        :param agent: Agent object.
        :return:
        """
        raise NotImplementedError(
            "Trainer.utility(agent) must be overridden in sub-class!"
        )

    def target_function(self, agent: Agent) -> float:
        """
        The target function to be minimized.

        :param agent:
        :return:
        """
        return -self.utility(agent)

    def add_agent_property(self, agent_list_name: str, prop: str):
        """
        Add a property of agent to be recorded.

        :param agent_list_name: Name of agent list
        :param prop: Property name of agent
        """
        self.container_manager.get_agent_container(
            agent_list_name
        ).recorded_properties.append(prop)

    def add_environment_property(self, prop: str):
        """
        Add a property of environment to be recorded.
        
        :param prop: Property name of environment.
        """
        assert prop not in self.environment_properties
        self.environment_properties.append(prop)

    def generate_scenarios(self):
        """
        Generate Scenarios for trainer

        :return: A list of scenario objects.
        """
        assert self.data_loader is not None
        return self.data_loader.generate_scenarios("trainer")

    def generate_trainer_params_list(
            self, trainer_scenario_cls: Type[GATrainerParams]
    ) -> List[GATrainerParams]:
        """
        Generate Trainer Parameters.

        :return: A list of trainer parameters.
        """
        trainer_params_table = self.get_dataframe("trainer_params_scenarios")
        assert isinstance(
            trainer_params_table, pd.DataFrame
        ), "No learning scenarios table specified!"

        trainer_params_raw_list = trainer_params_table.to_dict(orient="records")

        trainer_params_list = []
        for trainer_params_raw in trainer_params_raw_list:
            trainer_params = trainer_scenario_cls.from_dataframe_record(
                trainer_params_raw
            )
            trainer_params_list.append(trainer_params)
        return trainer_params_list



# -*- coding:utf-8 -*-
import os

from Melodie import Config

config = Config(
    project_name="_NAME_",
    project_root=os.path.dirname(__file__),
    sqlite_folder="data/sqlite",
    excel_source_folder="data/excel_source",
    output_folder="data/output",
)



# -*- coding:utf-8 -*-
# @Time: 2022/11/23 21:33
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: database.py

from MelodieInfra.jsonobject import JsonObject, StringProperty

class DatabaseBasicRequest(JsonObject):
    connection_string = StringProperty(name="connectionString", required=True)

class DatabaseQueryRequest(DatabaseBasicRequest):
    sql = StringProperty(name="sql", required=True)




# -*- coding:utf-8 -*-
# @Time: 2022/11/23 22:03
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: web.py.py
import json
from dataclasses import dataclass
from enum import Enum
from typing import Any


class DataServiceState(int, Enum):
    SUCCESS = 0
    ERROR = 1


@dataclass
class DataServiceStatus:
    status: DataServiceState
    msg: str = ""
    data: Any = None

    def to_json(self):
        return json.dumps({"msg": self.msg, "data": self.data, "status": self.status})



import logging

from MelodieInfra import *

from .utils import *
from .boost.agent_list import BaseAgentContainer, AgentList, AgentDict
from .boost.basics import Environment, Element, Agent
from .boost.grid import Grid, Spot, GridAgent, AgentIDManager
from .calibrator import Calibrator
from .data_collector import DataCollector
from .data_loader import DataLoader, DataFrameInfo, MatrixInfo
from .model import Model
from .network import Edge, Network, NetworkAgent
from .scenario_manager import Scenario
from .simulator import Simulator
from .table_generator import DataFrameGenerator
from .trainer import Trainer
from .visualizer import *

logging.basicConfig(
    level=logging.INFO,
    format="%(filename)s:%(lineno)d %(levelname)s %(message)s",
    datefmt="%a, %d %b %Y %H:%M:%S",
)



import argparse

from MelodieInfra import MelodieExceptions

parser = argparse.ArgumentParser()
parser.add_argument(
    "action",
    help="""[action] for starting the management server;
""",
)


def check_args(action: str):
    if action not in {"studio", "info"}:
        raise MelodieExceptions.Program.Variable.VariableNotInSet(
            "Command-Line argument 'action'", action, {"studio"}
        )


args = parser.parse_args()
check_args(args.action)
if args.action == "studio":
    from MelodieStudio.main import studio_main

    studio_main()
elif args.action == "info":
    from Melodie import get_system_info

    get_system_info()



# -*- coding:utf-8 -*-
# @Time: 2022/12/13 22:32
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: model_static_inspector.py
import ast
import os
import sys
from typing import Dict, List, Optional, Type, TYPE_CHECKING, Union

import astunparse
import pprintast

if TYPE_CHECKING:
    from Melodie import Agent, DataCollector, Environment, Model


def get_model_ast(model: 'Type[Model]'):
    file = sys.modules[model.__module__].__file__
    with open(file, encoding='utf-8', errors='replace') as f:
        for node in ast.walk(ast.parse(f.read())):
            if isinstance(node, ast.ClassDef) and node.name == model.__name__:
                return node
    raise Exception('No classdef in such file!')


class ModelStructure:
    def __init__(self):
        self.agent_types: 'List[Type[Agent]]' = []
        self.environment_type: 'Optional[Type[Environment]]' = None
        self.data_collector_type: 'Optional[Type[DataCollector]]' = None

        self.env_structure: "Optional[ClassStructure]" = None
        self.agents_structure: "Optional[Dict[str, ClassStructure]]" = {}

    def add_agent_container_type(self, agent_container_ast: ast.Name):
        self.agent_types.append(agent_container_ast.id)

    def set_env_type(self, env_ast: ast.Name):
        self.environment_type = env_ast.id

    def set_data_collector_type(self, dc_ast: ast.Name):
        self.data_collector_type = dc_ast.id

    def __repr__(self):
        return f"<{self.__class__} {self.__dict__}>"


class EnvironmentStructure:
    def __init__(self):
        self.properties = []


class Attribute:
    def __init__(self, name, readonly: bool = False, attr_type=None):
        self.name: str = name
        self.readonly: bool = readonly
        self.type: str = attr_type

    def __repr__(self):
        return f"{self.name} <{'readonly ' if self.readonly else ''}property>"


class Function:
    def __init__(self, name: str):
        self.name = name


class ClassStructure:
    def __init__(self):
        self.name: str = ""
        self.bases = []
        self.attributes: Dict[str, Attribute] = {}
        self.methods: Dict[str, Function] = {}

    def __repr__(self):
        return f"<ClassStructure of class '{self.name}' extends {self.bases}, attributes: {self.attributes}," \
               f" methods: {self.methods} >"

    @staticmethod
    def from_cls_ast(cls_ast: ast.ClassDef) -> "ClassStructure":
        cls_structure = ClassStructure()
        cls_structure.name = cls_ast.name
        cls_structure.parse_inheritance(cls_ast)
        for node in ast.walk(cls_ast):
            if isinstance(node, (ast.Assign, ast.AnnAssign)):
                cls_structure.parse_assign(node)
            elif isinstance(node, (ast.FunctionDef)):
                cls_structure.parse_functiondef(node)
        print(cls_structure)
        return cls_structure

    def parse_inheritance(self, cls_ast: ast.ClassDef):
        for base in cls_ast.bases:
            self.bases.append(astunparse.unparse(base).strip())

    def parse_assign(self, node: Union[ast.Assign, ast.AnnAssign]):
        targets = node.targets if isinstance(node, ast.Assign) else [node.target]
        for target in targets:
            if isinstance(target, ast.Attribute) and isinstance(target.value,
                                                                ast.Name) and target.value.id == 'self':
                self.attributes[target.attr] = Attribute(target.attr)

    def parse_functiondef(self, node: ast.FunctionDef):
        if len(node.decorator_list) > 0:
            for decorator_ast in node.decorator_list:
                if isinstance(decorator_ast, ast.Name) and decorator_ast.id == 'property':
                    if node.name not in self.attributes:
                        self.attributes[node.name] = Attribute(node.name, True)
                        return
                elif isinstance(decorator_ast, ast.Attribute) and decorator_ast.attr == 'setter' and isinstance(
                        decorator_ast.value, ast.Name):
                    attr_name = decorator_ast.value.id
                    if attr_name in self.attributes:
                        self.attributes[attr_name].readonly = False
                    else:
                        self.attributes[attr_name] = Attribute(node.name)
                    return
        self.methods[node.name] = Function(node.name)


def scan_attributes(cls_ast: ast.ClassDef) -> ClassStructure:
    """
    Scan attributes in a class

    :param cls_ast:
    :return:
    """
    return ClassStructure.from_cls_ast(cls_ast)


def find_class_in_files(cls_name: str, files_dir: str):
    print(files_dir)
    for root, dirs, files in os.walk(files_dir):
        for file in files:
            if file.endswith(".py"):
                file_abs_path = os.path.join(root, file)
                print(file_abs_path)
                with open(file_abs_path) as f:
                    tree: ast.Module = ast.parse(f.read())
                    for item in tree.body:
                        if isinstance(item, ast.ClassDef) and item.name == cls_name:
                            return item
    raise Exception("Class Not found!")


def walk_model_ast(model_ast: ast.ClassDef, file: str):
    model_structure = ModelStructure()
    functions = {'self.create_agent_list': lambda args: model_structure.add_agent_container_type(args[0]),
                 'self.create_environment': lambda args: model_structure.set_env_type(args[0]),
                 'self.create_data_collector': lambda args: model_structure.set_data_collector_type(args[0])}

    for node in ast.walk(model_ast):
        if isinstance(node, ast.Call):
            # print(astunparse.unparse(node.func))
            key = astunparse.unparse(node.func).strip()
            if key in functions:
                # print(functions)
                print(key, node.args)
                functions[key](node.args)
    # print(model_structure)

    for agent_type in model_structure.agent_types:
        agent_cls_ast = find_class_in_files(agent_type, os.path.dirname(file))
        model_structure.agents_structure[agent_type] = scan_attributes(agent_cls_ast)

    env_structure = scan_attributes(find_class_in_files(model_structure.environment_type, os.path.dirname(file)))
    model_structure.env_structure = env_structure
    # model_structure.agents_structure



import base64
import json
import logging
import os.path
import queue
import sys
import threading
import time
from typing import List, Tuple

import cloudpickle
from rpyc import Service
from rpyc.utils.server import ThreadedServer
import subprocess

logger = logging.getLogger("ParallelManager-MainThread")


class Tasks:
    def __init__(self):
        self.config = {}
        self.task_queue = queue.Queue()
        self.result_queue = queue.Queue()

    def put_task(self, item):
        self.task_queue.put(item)

    def get_task(self):
        try:
            return self.task_queue.get(block=False)
        except queue.Empty:
            return None

    def put_result(self, item):
        self.result_queue.put(item)

    def get_result(self):
        return self.result_queue.get()

    def get_config(self):
        return self.config


tasks: Tasks = None


class ParallelManager:
    def __init__(self, cores: int, configs: Tuple):
        self.th_server = threading.Thread(target=self.run_server)
        self.processes: List[subprocess.Popen] = []
        self.cores = cores
        self.server: ThreadedServer = None
        self.set_tasks(configs)

    def set_tasks(self, config):
        global tasks
        tasks = Tasks()
        tasks.config = config

    def put_task(self, task):
        tasks.put_task(task)

    def get_result(self):
        return tasks.get_result()

    def run_server(self):
        try:
            self.server = ThreadedServer(
                service=TimeService, port=12233, auto_register=False
            )
            self.server.start()
        except OSError as e:
            import traceback
            traceback.print_exc()
            # print()
            raise Exception("Server Error Occurred, exiting...")

    def run(self, role: str):
        """
        Run subprocesses

        :param role: run as calibrator or trainer
        :return:
        """
        assert role in {"calibrator", "trainer"}
        self.th_server.setDaemon(True)
        self.th_server.start()
        for core_id in range(self.cores):
            p = subprocess.Popen(
                [
                    sys.executable,
                    os.path.join(os.path.dirname(__file__), "parallel_worker.py"),
                    "--core_id",
                    str(core_id),
                    "--workdirs",
                    json.dumps([os.getcwd()]),
                    "--role",
                    role,
                ]
            )
            self.processes.append(p)

    def close(self):
        """
        Close all subprocesses

        :return:
        """
        for p in self.processes:
            p.terminate()
            logger.info(f"terminated subworker {p}")
        if self.server is not None:
            self.server.close()
            logger.info("Server closed!")
        global tasks
        tasks = None


class TimeService(Service):
    def exposed_get_time(self):
        return time.ctime()  # time模块中的一个内置方法

    def exposed_get_task(self):
        return base64.b64encode(cloudpickle.dumps(tasks.get_task()))

    def exposed_put_result(self, result):
        loaded = cloudpickle.loads(base64.b64decode(result))
        tasks.put_result(loaded)

    def exposed_get_config(self):
        return json.dumps(tasks.get_config())



# -*- coding:utf-8 -*-

import logging
from typing import Callable, Union, TYPE_CHECKING, Optional, Dict, Any

import pandas as pd

from MelodieInfra import MelodieExceptions
from .utils import args_check
from .scenario_manager import Scenario

if TYPE_CHECKING:
    from .data_loader import DataLoader, DataFrameInfo

logger = logging.getLogger(__name__)


class DataFrameGenerator:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        new_df = self.gen_agent_param_table_each_scenario()
        self.df_loader.register_dataframe(
            self.df_info.df_name, new_df, self.df_info.columns
        )
        return

    def __init__(
            self,
            df_loader: "DataLoader",
            df_info: "DataFrameInfo",
            num_generator: Union[int, Callable[[Scenario], int]],
    ):
        """
        :param df_loader:
        :param df_info:
        :param num_generator
        """

        self.num_generator = self.convert_to_num_generator(num_generator)
        self.df_info = df_info
        self._self_incremental_value = -1
        self.df_loader = df_loader
        from Melodie.data_loader import DataLoader

        if not isinstance(self.df_loader, DataLoader):
            MelodieExceptions.Data.NoDataframeLoaderDefined()

        self._row_generator: Optional[Callable[[Scenario], Union[dict, object]]] = None

    def increment(self):
        """
        Get increment value.

        :return:
        """
        self._self_incremental_value += 1
        return self._self_incremental_value

    def reset_increment(self):
        """
        Reset increment

        :return:
        """
        self._self_incremental_value = -1

    def convert_to_num_generator(
            self, num_generator: Union[int, Callable[[Scenario], int]]
    ):
        if isinstance(num_generator, int):
            return lambda _: num_generator
        elif callable(num_generator):
            args_check(num_generator, 1)
            return num_generator
        else:
            raise TypeError

    def set_row_generator(
            self, row_generator: Callable[[Union[Any, Scenario]], Union[Dict[str, Any]]]
    ):
        """
        Set the geneator for each row. Every time the row_generator is called, this function
        returns a dict standing for one row.
        For example, if row_generator is called twice, and the return values are {"id": 0, "a": 1},
            and {"id": 1, "a": 2}. The generated table will be:
        | id | a |
        |----|---|
        | 0  | 1 |
        | 1  | 2 |

        :param row_generator:
        :return:
        """
        args_check(row_generator, 1)
        self._row_generator = row_generator

    def gen_agent_param_table_each_scenario(self):
        """

        :return:
        """
        scenarios = self.df_loader.manager.generate_scenarios()
        data_list = []
        for scenario in scenarios:
            data_list.extend(self.gen_agent_params(scenario))
            self.reset_increment()
        return pd.DataFrame(data_list)

    def gen_agent_params(self, scenario: Scenario):
        """

        :param scenario:
        :return:
        """
        data_list = []
        for agent_id in range(0, self.num_generator(scenario)):
            d = {"id_scenario": scenario.id, "id": agent_id}
            generated = self._row_generator(scenario)
            if isinstance(generated, dict):
                d.update(generated)
            elif not type(generated).__module__ == "__builtin__":
                d.update(generated.__dict__)
            else:
                raise TypeError(
                    f"Builtin type {type(generated)} (value: {generated}) cannot be converted to table row."
                )
            data_list.append(d)
        return data_list



# -*- coding:utf-8 -*-
# @Time: 2022/9/4 10:47
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: test_props_holder.py
import time
from Melodie.boost.basics import PropHolder, foreach, length, times


class A:
    def __init__(self):
        self.a = 0


M = 10000000


def df_getitem_speed():
    holder = PropHolder()
    t0 = time.time()
    foreach()
    print("df:", time.time() - t0)


def atest2():
    s = 0
    a = A()
    t0 = time.time()
    for i in range(times):
        for j in range(length):
            a.a
            a.a = 123
    print("test2", time.time() - t0)


df_getitem_speed()
atest2()  # atest2



# -*- coding:utf-8 -*-
# @Time: 2022/11/23 21:42
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: __init__.py.py

from .base import ExcelMeta
from .excel import ExcelWriteRequest, ExcelReadSheetRequest, ExcelReadSheetResponse



import sys
from typing import Optional

DOCS_OFFICIAL_SITE = "https://abm4all.github.io/Melodie/html"
DOCS_MIRROR_SITE = "http://abm4all.gitee.io/melodie/html/"

HELP_URL_TEMPLATE = f"""
=====================Melodie Troubleshooting Tips==============================
The trouble shooting documentation for this error can be found at:

{DOCS_OFFICIAL_SITE}__URL__

or mirror site:

{DOCS_MIRROR_SITE}__URL__
================================================================================
"""


class OSTroubleShooter:
    """
    This class is a handler to print a link towards the help page.

    """
    _instance: Optional['OSTroubleShooter'] = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(OSTroubleShooter, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        self.cases = [(lambda s: "address already in use" in s or "Errno 10048" in s,
                       "/troubleshooting.html#port-already-in-use")]

    def handle_exc(self, exc: BaseException, exit_program=True, traceback=True):
        """
        Handle exception

        :param exc:
        :param exit_program: Program exit after printing troubleshooting url.
        :param traceback: Print traceback
        :return:
        """
        s = str(exc)
        for case in self.cases:
            if case[0](s):
                if traceback:
                    import traceback
                    traceback.print_exc()

                self._print_help_url(case[1])

                if exit_program:
                    sys.exit(1)

    @staticmethod
    def _print_help_url(url: str):
        print(HELP_URL_TEMPLATE.replace("__URL__", url), file=sys.stderr)



import logging

import numpy as np
from sko.GA import GA

logger = logging.getLogger(__name__)


class MelodieGA(GA):
    def run(self, max_iter=None):
        self.max_iter = max_iter or self.max_iter
        best = []
        for i in range(self.max_iter):
            self.X = self.chrom2x(self.Chrom)
            self.Y = self.x2y()
            self.ranking()
            self.selection()
            self.crossover()
            self.mutation()

            # record the best ones
            generation_best_index = self.FitV.argmax()
            self.generation_best_X.append(self.X[generation_best_index, :])
            self.generation_best_Y.append(self.Y[generation_best_index])
            self.all_history_Y.append(self.Y)
            self.all_history_FitV.append(self.FitV)

            if self.early_stop:
                best.append(min(self.generation_best_Y))
                if len(best) >= self.early_stop:
                    if best.count(min(best)) == len(best):
                        break
                    else:
                        best.pop(0)

        global_best_index = np.array(self.generation_best_Y).argmin()
        self.best_x = self.generation_best_X[global_best_index]
        self.best_y = self.generation_best_Y[global_best_index]
        return self.best_x, self.best_y



import os

import setuptools

from distutils.extension import Extension


def get_include():
    import numpy as np

    return [np.get_include()]


try:
    from Cython.Distutils import build_ext
    from Cython.Compiler import Options
    import Cython
    import numpy
    import sys

    Options.annotate = True
    ext_modules = [
        Extension(
            "Melodie.boost._vectorize",
            ["Melodie/boost/_vectorize.pyx"],
            language="c++",
            # extra_compile_args=["-std=c++11"]
        ),
        Extension(
            "Melodie.boost._vectorize2d",
            ["Melodie/boost/_vectorize2d.pyx"],
            language="c++",
            # extra_compile_args=["-std=c++11"]
        ),
        Extension(
            "Melodie.boost.agent_list",
            ["Melodie/boost/agent_list.pyx"],
            language="c++",
            # extra_compile_args=["-std=c++11"]
        ),
        Extension(
            "Melodie.boost.grid",
            ["Melodie/boost/grid.pyx"],
            language="c++",
            # extra_compile_args=["-std=c++11"],
        ),
        # define_macros=[('CYTHON_TRACE', '1')]),
        Extension("Melodie.boost.fastrand", ["Melodie/boost/fastrand.pyx"]),
        Extension(
            "Melodie.boost.basics",
            ["Melodie/boost/basics.pyx"],
            language="c++",
        ),
    ]
    ext = ".pyx"
    jsonobj_ext_modules = [
        Extension('MelodieInfra.jsonobject.api', [
                  "MelodieInfra/jsonobject/api" + ext], ),
        Extension('MelodieInfra.jsonobject.base', [
                  "MelodieInfra/jsonobject/base" + ext], ),
        Extension('MelodieInfra.jsonobject.base_properties', [
                  "MelodieInfra/jsonobject/base_properties" + ext], ),
        Extension('MelodieInfra.jsonobject.containers', [
                  "MelodieInfra/jsonobject/containers" + ext], ),
        Extension('MelodieInfra.jsonobject.properties', [
                  "MelodieInfra/jsonobject/properties" + ext], ),
        Extension('MelodieInfra.jsonobject.utils', [
                  "MelodieInfra/jsonobject/utils" + ext], ),
    ]
    ext_modules = ext_modules + jsonobj_ext_modules
    # ext_modules = ext_modules+[
    #     Extension('MelodieInfra.container.intmap',
    #               [
    #                   'MelodieInfra/container/intmap.pyx',
    #               ])
    # ]
except:
    import traceback

    traceback.print_exc()
    ext_modules = None

    def build_ext(_):
        return print(
            "Cython was not installed. With cython you may get better peformance boost!"
        )

with open("Melodie/version.txt", "r", encoding="utf8") as fv:
    version = fv.read()

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Melodie",
    version=version,
    description="A general framework that can be used to establish agent-based models for specific uses.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SongminYu/Melodie",
    author="Songmin Yu",
    author_email="songmin.yu@isi.fraunhofer.de",
    license="BSD 3",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        # 'Programming Language :: Python :: 3.6',
        # 'Programming Language :: Python :: 3.7',
        # 'Programming Language :: Python :: 3.8',
        # 'Programming Language :: Python :: 3.9',
        # 'Programming Language :: Python :: 3.10',
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    project_urls={
        "Documentation": "http://docs.abm4all.com",
    },
    packages=setuptools.find_namespace_packages(
        include=["Melodie", "Melodie.*", "MelodieInfra", "MelodieInfra.*"]),
    install_requires=[
        "chardet",
        "numpy",
        "pandas",
        "matplotlib",
        "seaborn",
        "networkx",
        "openpyxl",
        "websockets",
        "sqlalchemy",
        "astunparse",
        "pprintast",
        "cloudpickle",
        "cython==3.0.0a10",
        "scikit-opt~=0.6",
        "rpyc",
    ],
    python_requires=">=3.7",
    entry_points={"console_scripts": ["Melodie=Melodie.scripts.scripts:cli"]},
    include_package_data=True,
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    include_dirs=get_include(),
)



from __future__ import unicode_literals
class DeleteNotAllowed(Exception):
    pass


class BadValueError(Exception):
    """raised when a value can't be validated or is required"""


class WrappingAttributeError(AttributeError):
    pass



import logging
from typing import Optional, Union, Type, List

import pandas as pd

from MelodieInfra import create_db_conn, DBConn, MelodieExceptions, show_prettified_warning, show_link
from .boost.agent_list import AgentList, BaseAgentContainer, AgentDict
from .boost.basics import Agent, Environment
from .boost.grid import Grid, Spot
from MelodieInfra.config.config import Config
from .data_collector import DataCollector

from .scenario_manager import Scenario
from .table_generator import DataFrameGenerator
from .network import Network, Edge
from .visualizer import Visualizer

logger = logging.getLogger(__name__)


class ModelRunRoutine:
    """
    A simple iterator for model run.


    When calling ``Model.iterator()`` method, a ModelRunRoutine object will be created, yielding an ``int``  value
    reprensenting the current number of step, ranging ``[0, max_step - 1]``.


    """

    def __init__(self, max_step: int, model: "Model"):
        self._current_step = -1
        self._max_step = max_step
        self.model: "Model" = model

    def __iter__(self):
        return self

    def __next__(self):
        if self._current_step >= self._max_step - 1:
            raise StopIteration
        self.model._visualizer_step(self._current_step)
        self._current_step += 1
        return self._current_step

    def __del__(self):
        """
        Remove circular reference before deletion

        :return:
        """
        self.model = None


class Model:
    def __init__(
            self,
            config: "Config",
            scenario: "Scenario",
            run_id_in_scenario: int = 0,
            visualizer: Visualizer = None,
    ):

        self.scenario = scenario
        self.config = config

        self.environment: Optional[Environment] = None
        self.data_collector: Optional[DataCollector] = None
        self.table_generator: Optional[DataFrameGenerator] = None
        self.run_id_in_scenario = run_id_in_scenario

        self.network = None
        self.visualizer: "Visualizer" = visualizer
        self.initialization_queue: List[
            Union[AgentList, Grid, Environment, DataCollector, Network]
        ] = []

    def __del__(self):
        """
        Remove circular reference before deletion

        :return:
        """
        self.visualizer = None

    def create(self):
        """
        An initialization method, which is called immediately right after the ``Model`` object is created.

        :return: None
        """
        pass

    def setup(self):
        """
        General method for model setup, which is called after ``Model.create()``

        :return: None
        """
        pass

    def create_db_conn(self) -> "DBConn":
        """
        Create a database connection with the project configuration.

        :return: DBConn object
        """
        return create_db_conn(self.config)

    def create_agent_list(
            self,
            agent_class: Type["Agent"],
    ):
        """
        Create an agent list object. A model could contain multiple ``AgentList``s.

        :param agent_class: The class of desired agent type.
        :return: Agentlist object created
        """
        return AgentList(agent_class, model=self)

    def create_environment(self, env_class: Type["Environment"]):
        """
        Create the environment of model. Notice that a model has only one environment.

        :param env_class:
        :return: Environment object created
        """
        env = env_class()
        env.model = self
        env.scenario = self.scenario
        self.initialization_queue.append(env)
        return env

    def create_grid(self, grid_cls: Type["Grid"] = None, spot_cls: Type["Spot"] = None):
        """
        Create a grid.

        :param grid_cls: The class of grid, ``Melodie.Grid`` by default.
        :param spot_cls: The class of spot, ``Melodie.Spot`` by default.
        :return: Grid object.
        """
        grid_cls = grid_cls if grid_cls is not None else Grid
        spot_cls = spot_cls if spot_cls is not None else Spot
        grid = grid_cls(spot_cls, self.scenario)
        self.initialization_queue.append(grid)
        return grid

    def create_network(
            self, network_cls: Type["Network"] = None, edge_cls: Type["Edge"] = None
    ):
        """
        Create the network of model.

        :param network_cls: The type of network object, ``Melodie.Network`` by default.
        :param edge_cls: The type of edge object, ``Melodie.Edge`` by default.
        :return: Network object created
        """
        if network_cls is None:
            network_cls = Network
        network = network_cls(edge_cls)
        self.initialization_queue.append(network)
        return network

    def create_data_collector(self, data_collector_cls: Type["DataCollector"]):
        """
        Create the data collector of model.

        :param data_collector_cls: The datacollector class, must be a custom class inheriting ``Melodie.DataCollector``.
        :return: Datacollector object created.
        """
        data_collector = data_collector_cls()
        data_collector.model = self
        data_collector.scenario = self.scenario
        self.initialization_queue.append(data_collector)
        return data_collector

    def create_agent_container(
            self,
            agent_class: Type["Agent"],
            initial_num: int,
            params_df: pd.DataFrame = None,
            container_type: str = "list",
    ) -> Union[AgentList, AgentDict]:
        """
        Create a container for agents.

        :param agent_class:
        :param initial_num: Initial number of agents
        :param params_df:   Pandas DataFrame
        :param container_type: a str, "list" or "dict"
        :return: Agent container created
        """
        from Melodie import AgentList

        agent_container_class: Union[Type[AgentList], Type[AgentDict], None]
        if container_type == "list":
            agent_container_class = AgentList
        elif container_type == "dict":
            agent_container_class = AgentDict
        else:
            raise NotImplementedError(
                f"Container type '{container_type}' is not valid!"
            )

        container = agent_container_class(agent_class, model=self)
        if params_df is not None:
            container.set_properties(params_df)
        else:
            show_prettified_warning(
                f"No dataframe set for the {agent_container_class.__name__}.\n\t"
                + show_link()
            )
        self.initialization_queue.append(container)
        return container

    def _check_agent_containers(self):
        """
        Check the agent agent_lists in the model.
        Check list is:
        - Each agent, no matter which container it was in, should have a unique id.

        :return: None
        """
        for prop_name, prop in self.__dict__.items():
            if isinstance(prop, BaseAgentContainer):
                all_ids = prop.all_agent_ids()
                if len(set(all_ids)) < len(all_ids):
                    raise MelodieExceptions.Agents.AgentIDConflict(prop_name, all_ids)

    def run(self):
        """
        Model run. Be sure to inherit this method on your model.

        :return: None
        """
        pass

    def iterator(self, period_num: int):
        """
        Return an iterator which iterates from `0` to `period_num-1`. In each iteration, the iterator updates the
        visualizer if it exists.

        :param period_num: How many periods will this model run.
        :return: None
        """
        return ModelRunRoutine(period_num, self)

    def _visualizer_step(self, current_step: int):
        """
        If visualizer is defined, make it step.

        :param current_step:
        :return:
        """
        if (self.visualizer is not None) and (current_step > 0):
            self.visualizer.step(current_step)

    def init_visualize(self):
        """
        Be sure to implement it if you would like to use visualizer.

        :return:
        """
        # raise NotImplementedError

    def _setup(self):
        """
        Wrapper of setup()

        :return:
        """
        self.create()
        self.setup()
        for component_to_init in self.initialization_queue:
            component_to_init._setup()
